import functools
import math
from typing import Any, Callable, TypedDict

import numpy as np
from OCP.Geom import Geom_BSplineCurve, Geom_Curve
from OCP.gp import gp_Pnt, gp_Vec
from scipy.optimize import OptimizeResult, minimize_scalar

from .misc import (
    Standard_Real,
    clone_bspline,
    math_BFGS,
    math_MultipleVarFunctionWithGradient,
    math_Vector,
)

# --- Helper functions and classes from C++ ---


# def maxval(v1, v2):
#     return v1 if v1 > v2 else v2


# def sqr(v):
#     return v * v


def minCoords(p1: gp_Pnt, p2: gp_Pnt) -> gp_Pnt:
    return gp_Pnt(min(p1.X(), p2.X()), min(p1.Y(), p2.Y()), min(p1.Z(), p2.Z()))


def maxCoords(p1: gp_Pnt, p2: gp_Pnt) -> gp_Pnt:
    return gp_Pnt(max(p1.X(), p2.X()), max(p1.Y(), p2.Y()), max(p1.Z(), p2.Z()))


class Intervall:
    def __init__(self, mmin: float, mmax: float):
        self.min = mmin
        self.max = mmax

    def __eq__(self, other):
        EPS = 1e-15
        return abs(self.min - other.min) < EPS and abs(self.max - other.max) < EPS


@functools.lru_cache(maxsize=1024)
def _get_low_high_from_bspline(curve: Geom_BSplineCurve):
    np_array = np.zeros((curve.NbPoles(), 3), dtype=np.float64)
    for i in range(1, curve.NbPoles() + 1):
        p = curve.Pole(i)
        np_array[i - 1] = [
            p.X(),
            p.Y(),
            p.Z(),
        ]
    temp = np.min(np_array, axis=0)
    low = gp_Pnt(temp[0], temp[1], temp[2])
    temp = np.max(np_array, axis=0)
    high = gp_Pnt(temp[0], temp[1], temp[2])
    return low, high


class BoundingBox:
    def __init__(self, curve_or_other: "Geom_BSplineCurve | BoundingBox"):
        if isinstance(curve_or_other, Geom_BSplineCurve):
            curve = curve_or_other
            self.range = Intervall(curve.FirstParameter(), curve.LastParameter())
            self.low, self.high = _get_low_high_from_bspline(curve)
        else:
            other = curve_or_other
            self.range = Intervall(other.range.min, other.range.max)
            self.low = gp_Pnt(other.low.XYZ())
            self.high = gp_Pnt(other.high.XYZ())

    def Intersects(self, other: "BoundingBox", eps: float) -> bool:
        min_coords = maxCoords(self.low, other.low)
        max_coords = minCoords(self.high, other.high)
        return (
            (min_coords.X() < max_coords.X() + eps)
            and (min_coords.Y() < max_coords.Y() + eps)
            and (min_coords.Z() < max_coords.Z() + eps)
        )

    def Merge(self, other: "BoundingBox") -> "BoundingBox":
        self.range.max = other.range.max
        self.high = maxCoords(self.high, other.high)
        self.low = minCoords(self.low, other.low)
        return self

    def __eq__(self, other):
        return self.range == other.range


def curvature(curve: Geom_BSplineCurve) -> float:
    len_curve = curve.Pole(1).Distance(curve.Pole(curve.NbPoles()))
    total = 0.0
    for i in range(1, curve.NbPoles()):
        p1 = curve.Pole(i)
        p2 = curve.Pole(i + 1)
        dist = p1.Distance(p2)
        total += dist

    if abs(len_curve) < 1e-15:
        return 1.0 if total < 1e-15 else 1e9
    return max(total / len_curve, 1.0)


class BoundingBoxPair:
    def __init__(self, b1: BoundingBox, b2: BoundingBox):
        self.b1 = b1
        self.b2 = b2


class BSplineAlgorithms:
    @staticmethod
    def trimCurve(
        curve: Geom_BSplineCurve, first: float, last: float
    ) -> Geom_BSplineCurve:
        new_curve = clone_bspline(curve)
        new_curve.Segment(first, last)
        return new_curve


# --- Main function and classes to translate ---


# math_MultipleVarFunctionWithGradient
class CurveCurveDistanceObjective(math_MultipleVarFunctionWithGradient):
    def __init__(self, c1: Geom_Curve, c2: Geom_Curve):
        # Call the parent constructor without arguments, as math_MultipleVarFunction has no __init__
        super().__init__()
        self.m_c1 = c1
        self.m_c2 = c2

    def NbVariables(self) -> int:
        return 2

    def Value(self, X: math_Vector, F: Standard_Real) -> bool:
        G = math_Vector(1, 2)
        self.Values(X, F, G)
        return True  # The original C++ returns true, and F is an output parameter

    def Gradient(
        self, X: math_Vector, G: math_Vector
    ) -> bool:  # Gradient also takes G as an output parameter
        F_val = Standard_Real(0.0)  # Create a temporary Standard_Real for F
        self.Values(X, F_val, G)
        return True

    # C++ uses 0.5 * (math.sin(z) + 1) as the activate function to limit the u, v value to [0, 1]
    # But it is found that the optimization does not converge in some rare cases.
    # Here we use f(z)=z as the activate function which seems more stable for optimization.
    @staticmethod
    def activate(z: float) -> float:
        return z

    @staticmethod
    def d_activate(z: float) -> float:
        return 1

    # @staticmethod
    # def activate(z: float) -> float:
    #     return 0.5 * math.sin(z) + 0.5

    # @staticmethod
    # def d_activate(z: float) -> float:
    #     return 0.5 * math.cos(z)

    # @staticmethod
    # def activate(z: float) -> float:
    #     return 1.25 * math.atan(z) / math.pi + 0.5

    # @staticmethod
    # def d_activate(z: float) -> float:
    #     return 1.25 / (math.pi * (1.0 + z**2))

    def getUParam(self, x0: float) -> float:
        umin = self.m_c1.FirstParameter()
        umax = self.m_c1.LastParameter()
        return self.activate(x0) * (umax - umin) + umin

    def getVParam(self, x1: float) -> float:
        vmin = self.m_c2.FirstParameter()
        vmax = self.m_c2.LastParameter()
        return self.activate(x1) * (vmax - vmin) + vmin

    def d_getUParam(self, x0: float) -> float:
        umin = self.m_c1.FirstParameter()
        umax = self.m_c1.LastParameter()
        return self.d_activate(x0) * (umax - umin)

    def d_getVParam(self, x1: float) -> float:
        vmin = self.m_c2.FirstParameter()
        vmax = self.m_c2.LastParameter()
        return self.d_activate(x1) * (vmax - vmin)

    def Values(self, X: math_Vector, F: Standard_Real, G: math_Vector) -> bool:
        u = self.getUParam(X.Value(1))
        v = self.getVParam(X.Value(2))

        p1 = gp_Pnt(0, 0, 0)
        p2 = gp_Pnt(0, 0, 0)
        d1_vec = gp_Vec(0, 0, 0)
        d2_vec = gp_Vec(0, 0, 0)

        self.m_c1.D1(u, p1, d1_vec)
        self.m_c2.D1(v, p2, d2_vec)

        diff = gp_Vec(p2, p1)  # p1 - p2
        F.value = diff.SquareMagnitude()

        G.SetValue(
            1,
            2.0 * diff.Dot(d1_vec) * self.d_getUParam(X.Value(1)),
        )
        G.SetValue(
            2,
            -2.0 * diff.Dot(d2_vec) * self.d_getVParam(X.Value(2)),
        )

        return True


def find_split_point(curve: Geom_BSplineCurve):
    """
    Find the max inner multiplicity. If it is higher than 1, return the knot.
    otherwise return the mid point
    """
    mid_point = 0.5 * (curve.FirstParameter() + curve.LastParameter())
    if curve.IsCN(curve.Degree() - 1):
        return mid_point

    max_inner_mult_index = 2
    for i in range(3, curve.NbKnots()):
        if curve.Multiplicity(i) > curve.Multiplicity(max_inner_mult_index):
            max_inner_mult_index = i

    if curve.Multiplicity(max_inner_mult_index) > 1:
        return curve.Knot(max_inner_mult_index)

    return mid_point


def getRangesOfIntersection(
    curve1: Geom_BSplineCurve, curve2: Geom_BSplineCurve, tolerance: float
) -> list[BoundingBoxPair]:
    h1 = BoundingBox(curve1)
    h2 = BoundingBox(curve2)

    if not h1.Intersects(h2, tolerance):
        return []

    c1_curvature = curvature(curve1)
    c2_curvature = curvature(curve2)
    max_curvature = 1.0005

    if c1_curvature <= max_curvature and c2_curvature <= max_curvature:
        return [BoundingBoxPair(h1, h2)]

    curve1MidParm = find_split_point(curve1)
    curve2MidParm = find_split_point(curve2)

    results = []

    if c1_curvature > max_curvature and c2_curvature > max_curvature:
        c11 = BSplineAlgorithms.trimCurve(
            curve1, curve1.FirstParameter(), curve1MidParm
        )
        c12 = BSplineAlgorithms.trimCurve(curve1, curve1MidParm, curve1.LastParameter())

        c21 = BSplineAlgorithms.trimCurve(
            curve2, curve2.FirstParameter(), curve2MidParm
        )
        c22 = BSplineAlgorithms.trimCurve(curve2, curve2MidParm, curve2.LastParameter())

        results.extend(getRangesOfIntersection(c11, c21, tolerance))
        results.extend(getRangesOfIntersection(c11, c22, tolerance))
        results.extend(getRangesOfIntersection(c12, c21, tolerance))
        results.extend(getRangesOfIntersection(c12, c22, tolerance))

    elif c1_curvature <= max_curvature and max_curvature < c2_curvature:
        c21 = BSplineAlgorithms.trimCurve(
            curve2, curve2.FirstParameter(), curve2MidParm
        )
        c22 = BSplineAlgorithms.trimCurve(curve2, curve2MidParm, curve2.LastParameter())

        results.extend(getRangesOfIntersection(curve1, c21, tolerance))
        results.extend(getRangesOfIntersection(curve1, c22, tolerance))

    elif c2_curvature <= max_curvature and max_curvature < c1_curvature:
        c11 = BSplineAlgorithms.trimCurve(
            curve1, curve1.FirstParameter(), curve1MidParm
        )
        c12 = BSplineAlgorithms.trimCurve(curve1, curve1MidParm, curve1.LastParameter())

        results.extend(getRangesOfIntersection(c11, curve2, tolerance))
        results.extend(getRangesOfIntersection(c12, curve2, tolerance))

    return results


def replace_adjacent_in_list(
    avoid_u_values: list[float],
    list_obj: list[BoundingBox],
    is_adjacent_func: Callable[[list[float], BoundingBox, BoundingBox], bool],
    merge_func: Callable[[BoundingBox, BoundingBox], BoundingBox],
):
    i = 0
    while i < len(list_obj):
        next_i = i + 1
        if next_i >= len(list_obj):
            break

        if is_adjacent_func(avoid_u_values, list_obj[i], list_obj[next_i]):
            merged_val = merge_func(list_obj[i], list_obj[next_i])
            list_obj.pop(i)
            list_obj.pop(i)
            list_obj.insert(i, merged_val)
        else:
            i = next_i


def merge_boxes(b1: BoundingBox, b2: BoundingBox) -> BoundingBox:
    new_box = BoundingBox(b1)
    #             self.range = Intervall(other.range.min, other.range.max)
    #             self.low = gp_Pnt(other.low.XYZ())
    #             self.high = gp_Pnt(other.high.XYZ())
    new_box.Merge(b2)
    return new_box


# add two functions is_point_on_line_segment and line_line_intersection_3d
# to triage the intersection to avoid expensive minimizer
# and find good intial solution
# They are not in C++ code
def is_point_on_line_segment(
    point: gp_Pnt,
    segment_start: gp_Pnt,
    segment_end: gp_Pnt,
    max_curvature: float = 1.0005,
) -> tuple[float, bool]:
    """
    Checks if a 3D point lies on a 3D line segment which
    is the chord of a B-spline with the max_curvature.
    t is the parameter for the line segment in the range of [0, 1].
    """
    vec_segment = gp_Vec(segment_start, segment_end)
    vec_point_from_start = gp_Vec(segment_start, point)

    # t = (AP . AB) / |AB|^2
    line_length = vec_segment.Magnitude()
    if line_length < 1e-9:
        return 0, vec_point_from_start.SquareMagnitude() < 1e-9
    t = vec_point_from_start.Dot(vec_segment) / (line_length * line_length)

    possible_intersect = True

    # use a circular arc to estimate the max possible distance
    rel_sagitta = 0.6123724 * math.sqrt(max_curvature - 1)
    max_distance = 1.2 * rel_sagitta * line_length  # 1.2 for margin

    # check the range of t
    if t * line_length < -max_distance or t * line_length > line_length + max_distance:
        possible_intersect = False

    # Check the distance between the closest points
    closest_point_on_line = gp_Pnt(segment_start.XYZ())
    closest_point_on_line.BaryCenter(1 - t, segment_end, t)

    distance_to_line = point.Distance(closest_point_on_line)
    if distance_to_line > max_distance:
        possible_intersect = False

    return t, possible_intersect


def line_line_intersection_3d(
    p1_start: gp_Pnt,
    p1_end: gp_Pnt,
    p2_start: gp_Pnt,
    p2_end: gp_Pnt,
    max_curvature: float = 1.0005,
) -> tuple[float, float, bool]:
    """
    Finds the parameters (t, s) for the closest points on two 3D line segments which
    are the chords of possibly intersected two B-splines with the max_curvature.
    t is the parameter for line 1, s for line 2. Both are in the range of [0, 1].
    """
    u = gp_Vec(p1_start, p1_end)
    v = gp_Vec(p2_start, p2_end)
    w = gp_Vec(p1_start, p2_start)

    a = u.Dot(u)  # |u|^2
    b = u.Dot(v)
    c = v.Dot(v)  # |v|^2
    d = u.Dot(w)
    e = v.Dot(w)

    denom = a * c - b * b

    if abs(denom) < 1e-12:  # Lines are parallel or collinear
        t = 0.5
        s = 0.5
    else:
        t = (c * d - b * e) / denom
        s = (b * d - a * e) / denom

    possible_intersect = True

    # use a circular arc to estimate the max possible distance between two lines
    p1_mag = math.sqrt(a)
    p2_mag = math.sqrt(c)
    rel_sagitta = 0.6123724 * math.sqrt(max_curvature - 1)
    max_distance = 1.2 * rel_sagitta * (p1_mag + p2_mag)  # 1.2 for margin

    # check the range of t, s
    if t * p1_mag < -max_distance or t * p1_mag > p1_mag + max_distance:
        possible_intersect = False

    if s * p2_mag < -max_distance or s * p2_mag > p2_mag + max_distance:
        possible_intersect = False

    # Check the distance between the closest points
    closest_p1 = gp_Pnt(p1_start.XYZ())
    closest_p1.BaryCenter(1 - t, p1_end, t)
    closest_p2 = gp_Pnt(p2_start.XYZ())
    closest_p2.BaryCenter(1 - s, p2_end, s)

    distance = closest_p1.Distance(closest_p2)
    if distance > max_distance:
        possible_intersect = False

    return t, s, possible_intersect


class IntersectType(TypedDict):
    parmOnCurve1: float
    parmOnCurve2: float
    point: gp_Pnt


def IntersectBSplines(
    curve1: Geom_BSplineCurve, curve2: Geom_BSplineCurve, tolerance: float = 1e-5
) -> list[IntersectType]:
    hulls = getRangesOfIntersection(curve1, curve2, tolerance)

    curve1_ints: list[BoundingBox] = []
    curve2_ints: list[BoundingBox] = []
    for hull in hulls:
        curve1_ints.append(hull.b1)
        curve2_ints.append(hull.b2)

    curve1_ints.sort(key=lambda b: b.range.min)
    curve2_ints.sort(key=lambda b: b.range.min)

    unique_curve1_ints: list[BoundingBox] = []
    if curve1_ints:
        unique_curve1_ints.append(curve1_ints[0])
        for i in range(1, len(curve1_ints)):
            if not curve1_ints[i] == curve1_ints[i - 1]:
                unique_curve1_ints.append(curve1_ints[i])
    curve1_ints = unique_curve1_ints

    unique_curve2_ints: list[BoundingBox] = []
    if curve2_ints:
        unique_curve2_ints.append(curve2_ints[0])
        for i in range(1, len(curve2_ints)):
            if not curve2_ints[i] == curve2_ints[i - 1]:
                unique_curve2_ints.append(curve2_ints[i])
    curve2_ints = unique_curve2_ints

    def is_adjacent(
        avoid_u_values: list[float], b1: BoundingBox, b2: BoundingBox
    ) -> bool:
        EPS = 1e-15
        if (
            len(avoid_u_values) > 0
            and min(
                [
                    min(abs(b1.range.max - u), abs(b2.range.min - u))
                    for u in avoid_u_values
                ]
            )
            < EPS
        ):
            return False
        return abs(b1.range.max - b2.range.min) < EPS

    def create_avoid_u_values(curve: Geom_BSplineCurve):
        """
        We want to avoid merge adjacent boxes if the merge creates a kink
        """
        avoid_u_values: list[float] = []
        for i in range(2, curve.NbKnots()):
            if curve.Multiplicity(i) > 1:
                avoid_u_values.append(curve.Knot(i))
        return avoid_u_values

    replace_adjacent_in_list(
        create_avoid_u_values(curve1), curve1_ints, is_adjacent, merge_boxes
    )
    replace_adjacent_in_list(
        create_avoid_u_values(curve2), curve2_ints, is_adjacent, merge_boxes
    )

    intersectionCandidates: list[BoundingBoxPair] = []
    for b1 in curve1_ints:
        for b2 in curve2_ints:
            if b1.Intersects(b2, tolerance):
                intersectionCandidates.append(BoundingBoxPair(b1, b2))

    results: list[IntersectType] = []

    def append_result(results: list[IntersectType], result: IntersectType):
        for r in results:
            if (
                abs(r["parmOnCurve1"] - result["parmOnCurve1"]) < tolerance
                and abs(r["parmOnCurve2"] - result["parmOnCurve2"]) < tolerance
            ):
                return
        results.append(result)

    _tolerance_distance = max(1e-10, min(1e-5, tolerance))

    for boxes in intersectionCandidates:
        c1 = BSplineAlgorithms.trimCurve(curve1, boxes.b1.range.min, boxes.b1.range.max)
        c2 = BSplineAlgorithms.trimCurve(curve2, boxes.b2.range.min, boxes.b2.range.max)

        # When the intersection is at the endpoints, it is hard for 2d minimizer to converge.
        # So we check the endpoints first before using the 2d minimizer
        # C++ code does not have this checking.

        def get_mid_point(u: float, v: float):
            p1 = c1.Value(u)
            p2 = c2.Value(v)
            p1.BaryCenter(1.0, p2, 1.0)
            return p1

        # check if cx1.Value(param1) is on cx2
        def check_point_on_curve2(
            cx1: Geom_BSplineCurve, param1: float, cx2: Geom_BSplineCurve
        ):
            v = cx2.FirstParameter()
            cx2_first_point = cx2.Value(v)
            if cx1.Value(param1).Distance(cx2_first_point) < _tolerance_distance:
                return param1, v

            v = cx2.LastParameter()
            cx2_last_point = cx2.Value(v)
            if cx1.Value(param1).Distance(cx2_last_point) < _tolerance_distance:
                return param1, v

            t, possible_intersect = is_point_on_line_segment(
                cx1.Value(param1),
                cx2_first_point,
                cx2_last_point,
                max(curvature(cx1), curvature(cx2)),
            )
            if not possible_intersect:
                return None, None

            res = minimize_scalar(
                lambda v: cx1.Value(param1).SquareDistance(cx2.Value(v)),
                bounds=(cx2.FirstParameter(), cx2.LastParameter()),
                method="Bounded",
                options={"xatol": 0.1 * _tolerance_distance * _tolerance_distance},
            )
            if res.success and res.fun < _tolerance_distance * _tolerance_distance:  # type: ignore
                return param1, res.x  # type: ignore
            else:
                return None, None

        u, v = check_point_on_curve2(c1, c1.FirstParameter(), c2)
        if u is not None and v is not None:
            result: IntersectType = {
                "parmOnCurve1": u,
                "parmOnCurve2": v,
                "point": get_mid_point(u, v),  # c1.Value(c1.FirstParameter())
            }
            append_result(results, result)
            continue

        u, v = check_point_on_curve2(c1, c1.LastParameter(), c2)
        if u is not None and v is not None:
            result: IntersectType = {
                "parmOnCurve1": u,
                "parmOnCurve2": v,
                "point": get_mid_point(u, v),  # c1.Value(c1.LastParameter())
            }
            append_result(results, result)
            continue

        v, u = check_point_on_curve2(c2, c2.FirstParameter(), c1)
        if u is not None and v is not None:
            result: IntersectType = {
                "parmOnCurve1": u,
                "parmOnCurve2": v,
                "point": get_mid_point(u, v),  # c2.Value(c2.FirstParameter())
            }
            append_result(results, result)
            continue

        v, u = check_point_on_curve2(c2, c2.LastParameter(), c1)
        if u is not None and v is not None:
            result: IntersectType = {
                "parmOnCurve1": u,
                "parmOnCurve2": v,
                "point": get_mid_point(u, v),  # c2.Value(c2.LastParameter())
            }
            append_result(results, result)
            continue

        # C++ code starts from here. Hence, it only uses 2d minimizer without first checking intersection at endpoints
        obj = CurveCurveDistanceObjective(c1, c2)

        # Try to find a good initial guess using straight-line analytic solution
        p1_start = c1.Value(c1.FirstParameter())
        p1_end = c1.Value(c1.LastParameter())
        p2_start = c2.Value(c2.FirstParameter())
        p2_end = c2.Value(c2.LastParameter())

        t_closest, s_closest, possible_intersect = line_line_intersection_3d(
            p1_start,
            p1_end,
            p2_start,
            p2_end,
            max(curvature(c1), curvature(c2)),
        )

        if not possible_intersect:
            continue

        # Use custom math_Vector class for optimizer inputs
        guess = math_Vector(1, 2)
        guess.SetValue(1, t_closest)
        guess.SetValue(2, s_closest)

        # C++ uses 1e-10 as the aTolenrance for math_FRPR which seems to be an error
        for i in range(2):
            converged = math_BFGS(obj, guess, _tolerance_distance * _tolerance_distance)
            if converged:
                break
            # print(f"Warning: `IntersectBSplines` not converge {i+1} time")
            # print(f'boxes.b1.range=[{boxes.b1.range.min}, {boxes.b1.range.max}]')
            # print(f'boxes.b2.range=[{boxes.b2.range.min}, {boxes.b2.range.max}]')
            # print(f'tolerance={tolerance}')

        u = obj.getUParam(guess.Value(1))
        v = obj.getVParam(guess.Value(2))

        if (
            u < min(c1.FirstParameter(), c1.LastParameter()) - _tolerance_distance
            or u > max(c1.FirstParameter(), c1.LastParameter()) + _tolerance_distance
        ):
            continue

        if (
            v < min(c2.FirstParameter(), c2.LastParameter()) - _tolerance_distance
            or v > max(c2.FirstParameter(), c2.LastParameter()) + _tolerance_distance
        ):
            continue

        distance = c1.Value(u).Distance(c2.Value(v))
        # print(f'distance={distance}, _tolerance_distance={_tolerance_distance}, tolerance={tolerance}')

        if distance < tolerance:
            result: IntersectType = {
                "parmOnCurve1": u,
                "parmOnCurve2": v,
                "point": get_mid_point(u, v),
            }
            append_result(results, result)

    return results


# --- Main function call example (for testing purposes) ---
if __name__ == "__main__":
    pass



import numpy as np
import pytest
import sys
import os
import math

from OCP.gp import gp_Pnt, gp_Vec, gp_XYZ
from OCP.Geom import Geom_Curve, Geom_BSplineCurve
from OCP.TColgp import TColgp_Array1OfPnt
from OCP.GeomAPI import GeomAPI_PointsToBSpline

# Add the parent directory to the path to import the module
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from src_py.ocp_gordon.internal.misc import math_Vector, Standard_Real, math_MultipleVarFunctionWithGradient, math_BFGS

class CurveCurveDistanceObjectiveArcTan(math_MultipleVarFunctionWithGradient):
    def __init__(self, c1: Geom_Curve, c2: Geom_Curve):
        # Call the parent constructor without arguments, as math_MultipleVarFunction has no __init__
        super().__init__()
        self.m_c1 = c1
        self.m_c2 = c2

    def NbVariables(self) -> int:
        return 2

    def Value(self, X: math_Vector, F: Standard_Real) -> bool:
        G = math_Vector(1, 2)
        self.Values(X, F, G)
        return True # The original C++ returns true, and F is an output parameter

    def Gradient(self, X: math_Vector, G: math_Vector) -> bool: # Gradient also takes G as an output parameter
        F_val = Standard_Real(0.0) # Create a temporary Standard_Real for F
        self.Values(X, F_val, G)
        return True

    # @staticmethod
    # def activate(z: float) -> float:
    #     return 1.5 * math.atan(z) / math.pi + 0.5

    # @staticmethod
    # def d_activate(z: float) -> float:
    #     return 1.5 / (math.pi * (1.0 + z**2))

    @staticmethod
    def activate(z: float) -> float:
        return z #0.6 * math.sin(z) + 0.5

    @staticmethod
    def d_activate(z: float) -> float:
        return 1 #0.6 * math.cos(z)

    def getUParam(self, x0: float) -> float:
        umin = self.m_c1.FirstParameter()
        umax = self.m_c1.LastParameter()
        return self.activate(x0) * (umax - umin) + umin

    def getVParam(self, x1: float) -> float:
        vmin = self.m_c2.FirstParameter()
        vmax = self.m_c2.LastParameter()
        return self.activate(x1) * (vmax - vmin) + vmin

    def d_getUParam(self, x0: float) -> float:
        umin = self.m_c1.FirstParameter()
        umax = self.m_c1.LastParameter()
        return self.d_activate(x0) * (umax - umin)

    def d_getVParam(self, x1: float) -> float:
        vmin = self.m_c2.FirstParameter()
        vmax = self.m_c2.LastParameter()
        return self.d_activate(x1) * (vmax - vmin)

    def Values(self, X: math_Vector, F: Standard_Real, G: math_Vector) -> bool:
        u = self.getUParam(X.Value(1))
        v = self.getVParam(X.Value(2))

        p1 = gp_Pnt(0,0,0)
        p2 = gp_Pnt(0,0,0)
        d1_vec = gp_Vec(0,0,0)
        d2_vec = gp_Vec(0,0,0)

        self.m_c1.D1(u, p1, d1_vec)
        self.m_c2.D1(v, p2, d2_vec)

        diff = gp_Vec(p1.X() - p2.X(), p1.Y() - p2.Y(), p1.Z() - p2.Z())
        F.value = diff.SquareMagnitude()
        
        G.SetValue(1, 2. * diff.Dot(d1_vec) * (self.m_c1.LastParameter() - self.m_c1.FirstParameter()) * self.d_getUParam(X.Value(1)))
        G.SetValue(2, -2. * diff.Dot(d2_vec) * (self.m_c2.LastParameter() - self.m_c2.FirstParameter()) * self.d_getVParam(X.Value(2)))

        return True

class CurveCurveDistanceObjective(math_MultipleVarFunctionWithGradient):
    def __init__(self, c1: Geom_Curve, c2: Geom_Curve):
        # Call the parent constructor without arguments, as math_MultipleVarFunction has no __init__
        super().__init__()
        self.m_c1 = c1
        self.m_c2 = c2

    def NbVariables(self) -> int:
        return 2

    def Value(self, X: math_Vector, F: Standard_Real) -> bool:
        G = math_Vector(1, 2)
        self.Values(X, F, G)
        return True # The original C++ returns true, and F is an output parameter

    def Gradient(self, X: math_Vector, G: math_Vector) -> bool: # Gradient also takes G as an output parameter
        F_val = Standard_Real(0.0) # Create a temporary Standard_Real for F
        self.Values(X, F_val, G)
        return True

    @staticmethod
    def activate(z: float) -> float:
        return 0.5 * (math.sin(z) + 1.)

    @staticmethod
    def d_activate(z: float) -> float:
        return 0.5 * math.cos(z)

    def getUParam(self, x0: float) -> float:
        umin = self.m_c1.FirstParameter()
        umax = self.m_c1.LastParameter()
        return self.activate(x0) * (umax - umin) + umin

    def getVParam(self, x1: float) -> float:
        vmin = self.m_c2.FirstParameter()
        vmax = self.m_c2.LastParameter()
        return self.activate(x1) * (vmax - vmin) + vmin

    def d_getUParam(self, x0: float) -> float:
        umin = self.m_c1.FirstParameter()
        umax = self.m_c1.LastParameter()
        return self.d_activate(x0) * (umax - umin)

    def d_getVParam(self, x1: float) -> float:
        vmin = self.m_c2.FirstParameter()
        vmax = self.m_c2.LastParameter()
        return self.d_activate(x1) * (vmax - vmin)

    def Values(self, X: math_Vector, F: Standard_Real, G: math_Vector) -> bool:
        u = self.getUParam(X.Value(1))
        v = self.getVParam(X.Value(2))

        p1 = gp_Pnt(0,0,0)
        p2 = gp_Pnt(0,0,0)
        d1_vec = gp_Vec(0,0,0)
        d2_vec = gp_Vec(0,0,0)

        self.m_c1.D1(u, p1, d1_vec)
        self.m_c2.D1(v, p2, d2_vec)

        diff = gp_Vec(p1.X() - p2.X(), p1.Y() - p2.Y(), p1.Z() - p2.Z())
        F.value = diff.SquareMagnitude()
        
        G.SetValue(1, 2. * diff.Dot(d1_vec) * (self.m_c1.LastParameter() - self.m_c1.FirstParameter()) * self.d_getUParam(X.Value(1)))
        G.SetValue(2, -2. * diff.Dot(d2_vec) * (self.m_c2.LastParameter() - self.m_c2.FirstParameter()) * self.d_getVParam(X.Value(2)))

        return True
    
class QuadraticFunction(math_MultipleVarFunctionWithGradient):
    def __init__(self):
        super().__init__()
        self._nb_variables = 2

    def NbVariables(self) -> int:
        return self._nb_variables

    def Values(self, X: math_Vector, F: Standard_Real, G: math_Vector) -> bool:
        if X.Length() != self._nb_variables:
            return False

        x = X.Value(1)
        y = X.Value(2)

        # Function value: f(x, y) = x^2 + y^2
        F.value = x**2 + y**2

        # Gradient: grad_f(x, y) = [2x, 2y]
        G.SetValue(1, 2 * x)
        G.SetValue(2, 2 * y)
        
        return True

def create_bspline_curve(points: list[gp_Pnt]):
    """
    Create a B-spline curve from a list of points using GeomAPI_PointsToBSpline.
    
    Args:
        points: List of gp_Pnt points
        
    Returns:
        Handle(Geom_BSplineCurve): Approximated B-spline curve
    """
    # Create a regular array
    n_points = len(points)
    array = TColgp_Array1OfPnt(1, n_points)
    
    # Fill the array with points (indexing starts at 1 in OCP)
    for i, point in enumerate(points, 1):
            array.SetValue(i, point)
    
    # Create the approximator with reasonable defaults
    # Parameters: points, min_degree, max_degree, continuity, tolerance
    approximator = GeomAPI_PointsToBSpline(array)
    return approximator.Curve()

def create_test_curves():
    """
    Create test curves that form a proper intersecting network for Gordon surface interpolation.
    
    Returns:
        Tuple of (profiles, guides) - lists of B-spline curves that properly intersect
    """
    profiles: list[Geom_BSplineCurve] = []
    guides: list[Geom_BSplineCurve] = []
    
    # Define grid parameters
    num_profiles = 3
    num_guides = 4
    u_range = 8.0  # Range in u-direction (profiles)
    v_range = 5.0  # Range in v-direction (guides)
    
    # Create intersection points grid
    # This defines where profiles and guides should intersect
    intersection_points = np.zeros((num_profiles, num_guides, 3))
    
    for i in range(num_profiles):
        for j in range(num_guides):
            # Create a grid of points with some variation for a more interesting surface
            u = i * u_range / (num_profiles - 1) if num_profiles > 1 else 0
            v = j * v_range / (num_guides - 1) if num_guides > 1 else 0
            
            # Add some 3D variation to make the surface more interesting
            z = 0.5 * np.sin(u * 0.5) * np.cos(v * 0.5)
            
            intersection_points[i, j] = [u, v, z]
    
    # Create profile curves (u-direction)
    for i in range(num_profiles):
        points: list[gp_Pnt] = []
        
        # Each profile curve goes through all guide intersection points at this profile index
        for j in range(num_guides):
            x, y, z = intersection_points[i, j]
            points.append(gp_Pnt(x, y, z))
        
        # Create B-spline curve through these points
        bspline_curve = create_bspline_curve(points)
        profiles.append(bspline_curve)
    
    # Create guide curves (v-direction)
    for j in range(num_guides):
        points: list[gp_Pnt] = []
        
        # Each guide curve goes through all profile intersection points at this guide index
        for i in range(num_profiles):
            x, y, z = intersection_points[i, j]
            points.append(gp_Pnt(x, y, z))
        
        # Create B-spline curve through these points
        bspline_curve = create_bspline_curve(points)
        guides.append(bspline_curve)
    
    return profiles, guides

def test_math_bfgs_CurveCurveDistanceObjective():
    """
    Tests the math_BFGS function with test_math_bfgs_CurveCurveDistanceObjective.
    The expected minimum is at (0,0).
    """
    print("\nTesting math_BFGS with test_math_bfgs_CurveCurveDistanceObjective")

    # Initial guess
    initial_x = math_Vector(1, 2)
    initial_x.SetValue(1, 0)
    initial_x.SetValue(2, 0)

    profiles, guides = create_test_curves()

    # curve1 = profiles[0]
    # curve2 = guides[0]
    # curve1.Segment(0, 1.0)
    # curve2.Segment(0, 0.125)

    curve1 = profiles[1]
    curve2 = guides[0]
    curve1.Segment(0, 0.0625)
    curve2.Segment(0.425, 0.5)

    # curve1 = profiles[2]
    # curve2 = guides[0]
    # curve1.Segment(0, 0.125)
    # curve2.Segment(0.875, 1.0)

    # func = CurveCurveDistanceObjective(profiles[0], guides[0])
    func = CurveCurveDistanceObjectiveArcTan(curve1, curve2)

    # Set tolerance and max iterations
    tolerance = 1e-10
    max_iterations = 800

    print(f"Initial X: ({initial_x.Value(1)}, {initial_x.Value(2)})")
    initial_F = Standard_Real(0.0)
    initial_G = math_Vector(1, 2)
    func.Values(initial_x, initial_F, initial_G)
    print(f"Initial F: {initial_F.value}")
    print(f"Initial G: ({initial_G.Value(1)}, {initial_G.Value(2)})")

    # Run FRPR
    success = math_BFGS(func, initial_x, tolerance, max_iterations)

    def get_point(p: gp_Pnt):
        return [p.X(), p.Y(), p.Z()]
    
    u = func.getUParam(initial_x.Value(1))
    v = func.getVParam(initial_x.Value(2))
    print(f'u={u}, v={v}, curve1({u})={get_point(curve1.Value(u))}, curve2({v})={get_point(curve2.Value(v))}')
    # assert False

    final_F = Standard_Real(0.0)
    final_G = math_Vector(1, 2)

    print("\n--- Optimization Results ---")
    if success:
        print("Optimization successful!")
        print(f"Final X: ({initial_x.Value(1)}, {initial_x.Value(2)})")
        final_F = Standard_Real(0.0)
        final_G = math_Vector(1, 2)
        func.Values(initial_x, final_F, final_G)
        print(f"Final F: {final_F.value}")
        print(f"Final G: ({final_G.Value(1)}, {final_G.Value(2)})")
    # Assertions for pytest
    assert success, "Optimization should have succeeded"
    # assert abs(u) < 0.1, f"Final X[1] should be close to 0, got {u}"
    # assert abs(v) < 0.1, f"Final X[2] should be close to 0, got {v}"
    
    if success: # Only assert final F and G if optimization was successful
        assert abs(final_F.value) < tolerance, f"Final F should be close to 0, got {final_F.value}"
        # assert abs(final_G.Value(1)) < tolerance, f"Final G[1] should be close to 0, got {final_G.Value(1)}"
        # assert abs(final_G.Value(2)) < tolerance, f"Final G[2] should be close to 0, got {final_G.Value(2)}"

def test_math_bfgs_quadratic_function():
    """
    Tests the math_BFGS function with a simple quadratic function (f(x,y) = x^2 + y^2).
    The expected minimum is at (0,0).
    """
    print("\nTesting math_BFGS with QuadraticFunction (f(x,y) = x^2 + y^2)")

    # Initial guess
    initial_x = math_Vector(1, 2)
    initial_x.SetValue(1, 3.0)
    initial_x.SetValue(2, 4.0)

    # Create function instance
    quad_func = QuadraticFunction()

    # Set tolerance and max iterations
    tolerance = 1e-6
    max_iterations = 100

    print(f"Initial X: ({initial_x.Value(1)}, {initial_x.Value(2)})")
    initial_F = Standard_Real(0.0)
    initial_G = math_Vector(1, 2)
    quad_func.Values(initial_x, initial_F, initial_G)
    print(f"Initial F: {initial_F.value}")
    print(f"Initial G: ({initial_G.Value(1)}, {initial_G.Value(2)})")

    # Run FRPR
    success = math_BFGS(quad_func, initial_x, tolerance, max_iterations)

    final_F = Standard_Real(0.0)
    final_G = math_Vector(1, 2)

    print("\n--- Optimization Results ---")
    if success:
        print("Optimization successful!")
        print(f"Final X: ({initial_x.Value(1)}, {initial_x.Value(2)})")
        final_F = Standard_Real(0.0)
        final_G = math_Vector(1, 2)
        quad_func.Values(initial_x, final_F, final_G)
        print(f"Final F: {final_F.value}")
        print(f"Final G: ({final_G.Value(1)}, {final_G.Value(2)})")
    # Assertions for pytest
    assert success, "Optimization should have succeeded"
    assert abs(initial_x.Value(1)) < 0.1, f"Final X[1] should be close to 0, got {initial_x.Value(1)}"
    assert abs(initial_x.Value(2)) < 0.1, f"Final X[2] should be close to 0, got {initial_x.Value(2)}"
    
    if success: # Only assert final F and G if optimization was successful
        assert abs(final_F.value) < tolerance, f"Final F should be close to 0, got {final_F.value}"
        # assert abs(final_G.Value(1)) < tolerance, f"Final G[1] should be close to 0, got {final_G.Value(1)}"
        # assert abs(final_G.Value(2)) < tolerance, f"Final G[2] should be close to 0, got {final_G.Value(2)}"


if __name__ == "__main__":
    if 0:
        pytest.main([f'{__file__}::test_math_bfgs_CurveCurveDistanceObjective', "-v"])
    else:
        pytest.main([f'{__file__}', "-v"])



"""
Internal modules for Gordon curve interpolation.

This package contains the core implementation components for the
Gordon surface interpolation algorithm.
"""

from .bspline_algorithms import BSplineAlgorithms
from .intersect_bsplines import IntersectBSplines
from .gordon_surface_builder import GordonSurfaceBuilder
from .curve_network_sorter import CurveNetworkSorter
from .curves_to_surface import CurvesToSurface
from .points_to_bspline_interpolation import PointsToBSplineInterpolation
from .interpolate_curve_network import interpolate_curve_network

__all__ = [
    'BSplineAlgorithms',
    'IntersectBSplines',
    'GordonSurfaceBuilder',
    'CurveNetworkSorter',
    'CurvesToSurface',
    'PointsToBSplineInterpolation',
    'interpolate_curve_network'
]



"""
Simple example demonstrating Gordon curve interpolation with build123d.

This example shows how to use the Gordon surface interpolation
functionality with build123d's OCP integration.
"""
# %%
from typing import Union, List
from build123d import *  # type: ignore
import numpy as np
from ocp_vscode import show, Camera, set_defaults

# set_defaults(reset_camera=Camera.CENTER, helper_scale=5)
set_defaults(reset_camera=Camera.KEEP, helper_scale=5)

from build123d_face_ext import Face_ext

# %%

# simple sine wave surface
def create_test_curves1():
    """
    Create test curves that form a proper intersecting network for Gordon surface interpolation.
    
    Returns:
        Tuple of (profiles, guides) - lists of B-spline curves that properly intersect
    """
    profiles: list[Edge] = []
    guides: list[Edge] = []
    
    # Define grid parameters
    num_profiles = 8
    num_guides = 8
    u_range = 5.0  # Range in u-direction (profiles)
    v_range = 8.0  # Range in v-direction (guides)
    
    # Create intersection points grid
    # This defines where profiles and guides should intersect
    intersection_points = np.zeros((num_profiles, num_guides, 3))
    
    for i in range(num_profiles):
        for j in range(num_guides):
            # Create a grid of points with some variation for a more interesting surface
            u = i * u_range / (num_profiles - 1) if num_profiles > 1 else 0
            v = j * v_range / (num_guides - 1) if num_guides > 1 else 0
            
            # Add some 3D variation to make the surface more interesting
            z = 0.5 * np.sin(u * 0.5) * np.cos(v * 0.5)
            
            intersection_points[i, j] = [u, v, z]
    
    # Create profile curves (u-direction)
    for i in range(num_profiles):
        points: list[VectorLike] = []
        
        # Each profile curve goes through all guide intersection points at this profile index
        for j in range(num_guides):
            x, y, z = intersection_points[i, j]
            points.append((x, y, z))
        
        # Create B-spline curve through these points
        bspline_curve = Spline(points)
        profiles.append(bspline_curve)
    
    # Create guide curves (v-direction)
    for j in range(num_guides):
        points: list[VectorLike] = []
        
        # Each guide curve goes through all profile intersection points at this guide index
        for i in range(num_profiles):
            x, y, z = intersection_points[i, j]
            points.append((x, y, z))
        
        # Create B-spline curve through these points
        bspline_curve = Spline(points)
        guides.append(bspline_curve)
    
    return profiles, guides

# aircraft engine cowling
def create_test_curves2():
    """
    Create test curves that form a proper intersecting network for Gordon surface interpolation.
    
    Returns:
        Tuple of (profiles, guides) - lists of B-spline curves that properly intersect
    """    
    # Define grid parameters
    radius = 10
    length = 36

    outer = Spline([(0.8, 1), (1.1, 0.35), (1.0, 0)])
    inner = Spline([(0.9, 0), (0.85, 0.35), (0.7, 1)])
    num_points = 40
    points = *[outer@(i/num_points) for i in range(num_points+1)], *[inner@(i/num_points) for i in range(num_points+1)]
    points = [Vector(p.X * radius, p.Y * length) for p in points]

    guide1 = Spline(points)
    guides: List[Edge] = [Rot(0,i*90) * guide1 for i in range(4)] # type: ignore
    # show(guide1, *points, reset_camera=Camera.KEEP)

    def to_circle(v: Vector) -> Edge:
        return Pos(0, v.Y) * Rot(90) * CenterArc((0,0,0), 1, 0, 360).scale(v.X) # type: ignore
    
    profiles = [to_circle(guide1@0), to_circle(guide1@1)]
    
    return profiles, guides

# airliner fuselage
def create_test_curves3():
    profiles: list[Edge] = []
    guides: list[Edge] = []

    # Define points for the aircraft side profile in the YZ plane (X=0)
    # These points approximate the shape of an aircraft fuselage from nose to tail.
    # Y-coordinates are positive, Z-coordinates define the vertical profile.
    top_guide_points = [
        (0, 0, 1.47),
        (0, 1, 2.21),
        (0, 2.2, 2.94),
        (0, 3, 3.68),
        (0, 4, 4.56),
        (0, 5, 5.15),
        (0, 7, 5.70),
        (0, 10, 6.05),
        (0, 15, 6.25),
        (0, 30, 6.25),
        (0, 50, 6.21),
        (0, 60, 6.10),
        (0, 74.5, 4.96),
    ]

    bottom_guide_points = [
        (0, 0, -0.37),
        (0, 1, -0.92),
        (0, 2.2, -1.32),
        (0, 3, -1.54),
        (0, 4, -1.77),
        (0, 5, -1.91),
        (0, 7, -2.02),
        (0, 10, -2.10),
        (0, 15, -2.13),
        (0, 30, -2.10),
        (0, 48, -2.10),
        (0, 50, -2.06),
        (0, 52, -1.91),
        (0, 60, -0.55),
        (0, 68, 1.47),
        (0, 74.5, 3.49),
    ]

    guide1 = Spline(top_guide_points) # top guide
    guide2 = Spline(bottom_guide_points) # bottom guide
    
    # points = [Vector(*p) for p in top_guide_points]

    # create a circle passing p1 and p2 as diameter
    def circle_by_2_point(p1: Vector, p2: Vector):
        center = (p1 + p2) / 2
        loc1 = Location(center, (90 + Vector(0, 0, 1).get_signed_angle(p2 - p1), 0, 0))
        c1 = CenterArc(center=(0,0,0), radius=abs(p1-p2)/2, start_angle=0, arc_size=360)
        return c1.locate(loc1)

    profile_section_points = [0, 5, 10, 30, 50, top_guide_points[-1][1]]

    for section_point in profile_section_points:
        point1 = guide1.intersect(Plane((0, section_point, 0), z_dir=(0, -1, 0)))
        point2 = guide2.intersect(Plane((0, section_point, 0), z_dir=(0, -1, 0)))
        if point1 is not None and point2 is not None:
            vertex1 = point1.vertex()
            vertex2 = point2.vertex()
            if vertex1 is not None and vertex2 is not None:
                profiles.append(circle_by_2_point(vertex1.center(), vertex2.center()))

    guide_points = [0]
    guides = [guide1, guide2]
    for u in guide_points:
        guide3 = Spline([p@u for p in profiles])
        guide4 = guide3.mirror(Plane.YZ)
        guides.extend([guide3, guide4])
    
    # show(*points, *profiles, *guides)
    return profiles, guides

if __name__ == "__main__":
    """Main demonstration function."""
    print("Gordon Curve Interpolation Example")
    print("==================================")
    
    # Create test curves
    print("Creating test curves...")
    # profiles, guides = create_test_curves1()
    # profiles, guides = create_test_curves2()
    profiles, guides = create_test_curves3()
    # show(*profiles, *guides, reset_camera=Camera.KEEP)
    
    print(f"Created {len(profiles)} profile curves and {len(guides)} guide curves")
    
    # Perform Gordon interpolation
    print("Performing Gordon surface interpolation...")
    
    if 1:
        face10 = Face_ext.gordon_surface(
            profiles, guides, tolerance=3e-4
        )

        edge20 = (face10.edges() > Axis.Y)[0]
        face20 = Face.make_surface_patch([(edge20, face10, ContinuityLevel.C2)])

        edge30 = (face10.edges() < Axis.Y)[0]
        face30 = Face.make_surface_patch([(edge30, face10, ContinuityLevel.C0)])

        shell10 = Shell((face10, face20, face30))
        solid10 = Solid(shell10)

        show(solid10, *profiles, *guides)
        



# %%
# This file extends the class build123d.Face to include gorden_surface()

import json
from pathlib import Path
from build123d import VectorLike, Vector, Edge, Face, ShapeList  # type: ignore
from OCP.Geom import Geom_Curve, Geom_BSplineCurve, Geom_BSplineSurface
from OCP.GeomAPI import GeomAPI_PointsToBSpline
from OCP.GeomConvert import GeomConvert
from OCP.TColgp import TColgp_Array1OfPnt
from OCP.TColStd import TColStd_Array1OfReal, TColStd_Array1OfInteger
from OCP.BRepBuilderAPI import BRepBuilderAPI_MakeEdge, BRepBuilderAPI_MakeFace
from OCP.BRep import BRep_Tool
from OCP.TopoDS import TopoDS_Face
from OCP.gp import gp_Pnt
from OCP.Precision import Precision

# Import our Gordon interpolation module
import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

# from ocp_gordon import interpolate_curve_network, interpolate_curve_network_debug
from src_py.ocp_gordon import interpolate_curve_network, interpolate_curve_network_debug

# %%
def convert_bspline_to_edge(curve: Geom_Curve | Geom_BSplineCurve) -> Edge:
    make_edge = BRepBuilderAPI_MakeEdge(curve)
    ocp_edge = make_edge.Edge()
    return Edge(ocp_edge)

def convert_edge_to_bspline(edge: Edge) -> Geom_BSplineCurve:
    if edge.wrapped is None:
        raise ValueError("Edge cannot be empty")
    curve = BRep_Tool.Curve_s(edge.wrapped, 0, 1)
    return GeomConvert.CurveToBSplineCurve_s(curve)

def convert_bspline_surface_to_face(surface: Geom_BSplineSurface) -> Face:
    make_face = BRepBuilderAPI_MakeFace(surface, 1e-4) # Use a small tolerance
    ocp_face = make_face.Face()
    return Face(ocp_face)

def convert_face_to_bspline_surface(face: TopoDS_Face) -> Geom_BSplineSurface:
    """
    Converts a TopoDS_Face to a Geom_BSplineSurface.
    """
    # 1. Extract the underlying geometric surface from the TopoDS_Face
    geom_surface = BRep_Tool.Surface_s(face)

    # 2. Convert the Geom_Surface to a Geom_BSplineSurface
    bspline_surface = GeomConvert.SurfaceToBSplineSurface_s(geom_surface)

    return bspline_surface


class Face_ext(Face):
    """
    Extends build123d.Face with a gordon_surface method for interpolation.
    """

    @classmethod
    def gordon_surface(
        cls,
        profiles: list[Edge] | ShapeList[Edge],
        guides: list[Edge] | ShapeList[Edge],
        tolerance: float = 3e-4,
    ):
        """
        Creates a Gordon surface from a network of profile and guide curves.

        Args:
            profiles: A list or ShapeList of build123d.Edge objects representing profile curves.
            guides: A list or ShapeList of build123d.Edge objects representing guide curves.
            tolerance: Tolerance for surface creation and intersection calculations.

        Returns:
            A Face_ext object representing the interpolated Gordon surface.
        """
        ocp_profiles: list[Geom_Curve] = []
        ocp_guides: list[Geom_Curve] = []

        for edge in profiles:
            if edge.wrapped is None:
                raise ValueError("error")
            ocp_profiles.append(BRep_Tool.Curve_s(edge.wrapped, 0, 1))

        for edge in guides:
            if edge.wrapped is None:
                raise ValueError("error")
            ocp_guides.append(BRep_Tool.Curve_s(edge.wrapped, 0, 1))

        gordon_bspline_surface = interpolate_curve_network(
            ocp_profiles, ocp_guides, tolerance=tolerance
        )

        return cls(BRepBuilderAPI_MakeFace(gordon_bspline_surface, Precision.Confusion_s()).Face())

    # @classmethod
    # def gordon_surface_debug(
    #     cls,
    #     profiles: list[Edge] | ShapeList[Edge],
    #     guides: list[Edge] | ShapeList[Edge],
    #     tolerance: float = 1e-4,
    # ):
    #     """
    #     Creates a Gordon surface from a network of profile and guide curves.

    #     Args:
    #         profiles: A list or ShapeList of build123d.Edge objects representing profile curves.
    #         guides: A list or ShapeList of build123d.Edge objects representing guide curves.
    #         tolerance: Tolerance for surface creation and intersection calculations.

    #     Returns:
    #         A Face_ext object representing the interpolated Gordon surface.
    #     """
    #     ocp_profiles: list[Geom_Curve] = []
    #     ocp_guides: list[Geom_Curve] = []

    #     for edge in profiles:
    #         ocp_profiles.append(BRep_Tool.Curve_s(edge.wrapped, edge.param_at(0), edge.param_at(1)))

    #     for edge in guides:
    #         ocp_guides.append(BRep_Tool.Curve_s(edge.wrapped, edge.param_at(0), edge.param_at(1)))

    #     interpolator = interpolate_curve_network_debug(
    #         ocp_profiles, ocp_guides, tolerance=tolerance
    #     )

    #     # Create a Face from the OCP surface and then wrap it in Face_ext
    #     temp_face10 = convert_bspline_surface_to_face(interpolator.surface())
    #     temp_face20 = convert_bspline_surface_to_face(interpolator.surface_profiles())
    #     temp_face30 = convert_bspline_surface_to_face(interpolator.surface_guides())
    #     temp_face40 = convert_bspline_surface_to_face(interpolator.surface_intersections())
    #     return cls(temp_face10.wrapped), cls(temp_face20.wrapped), cls(temp_face30.wrapped), cls(temp_face40.wrapped)

    # def get_poles(self):
    #     ocp_surf: Geom_BSplineSurface = convert_face_to_bspline_surface(self.wrapped)
    #     poles: list[List[Vector]] = []
    #     for u in range(1, ocp_surf.NbUPoles()+1):
    #         poles.append([])
    #         for v in range(1, ocp_surf.NbVPoles()+1):
    #             pnt = ocp_surf.Pole(u, v)
    #             poles[-1].append(Vector(pnt))
    #     return poles
    
    # @staticmethod
    # def edge_to_spline(edge: Edge) -> Edge:
    #     curve = BRep_Tool.Curve_s(edge.wrapped, 0, 1)
    #     bspline = GeomConvert.CurveToBSplineCurve_s(curve)
    #     return convert_bspline_to_edge(bspline)


if __name__ == "__main__":
    pass



"""
Surface skinning from curves.

This module provides functionality to create surfaces by skinning
a set of compatible B-spline curves.
"""

from typing import List, Optional, Tuple

import numpy as np
from OCP.BSplCLib import BSplCLib  # For C++ BSplCLib equivalent
from OCP.Geom import (
    Geom_BSplineCurve,
    Geom_BSplineSurface,
    Geom_Curve,
    Geom_TrimmedCurve,
)
from OCP.GeomAPI import (  # Equivalent to Geom2dAPI_ProjectPointOnCurve for 3D
    GeomAPI_Interpolate,
)
from OCP.GeomConvert import GeomConvert
from OCP.gp import gp_Pnt
from OCP.Precision import Precision  # For C++ Precision equivalent
from OCP.TColgp import (
    TColgp_Array1OfPnt,
    TColgp_Array2OfPnt,
    TColgp_HArray1OfPnt,
    TColgp_HArray1OfPnt2d,
)
from OCP.TColStd import (
    TColStd_Array1OfInteger,
    TColStd_Array1OfReal,
    TColStd_Array2OfReal,
    TColStd_HArray1OfInteger,
    TColStd_HArray1OfReal,
)

from .bspline_algorithms import BSplineAlgorithms
from .error import ErrorCode, error  # Import ErrorCode
from .points_to_bspline_interpolation import (  # Import PointsToBSplineInterpolation
    PointsToBSplineInterpolation,
)


# Helper function to clamp BSpline (equivalent to C++ clampBSpline)
def clamp_bspline(curve: Geom_BSplineCurve):
    """
    Clamps a periodic B-spline curve by removing periodicity.
    Equivalent to C++ clampBSpline.
    """
    if not curve.IsPeriodic():
        return

    # C++ logic: SetNotPeriodic, trim to original range, convert back.
    curve.SetNotPeriodic()

    first_param = curve.FirstParameter()
    last_param = curve.LastParameter()

    # Create a new curve by trimming to the original parameter range
    trimmed_curve = Geom_TrimmedCurve(curve, first_param, last_param)

    # Convert back to BSplineCurve
    new_curve = GeomConvert.CurveToBSplineCurve_s(trimmed_curve)

    # Update the original curve handle (or return new_curve if preferred)
    # For in-place modification, we need to be careful with handles.
    # Let's assume we can modify the curve object directly or replace it.
    # In Python, modifying the object passed might not update the original if it's a copy.
    # However, Geom_BSplineCurve is likely a handle-based object in OCP.
    # For safety, let's return the new curve and let the caller assign it.
    # If the caller expects in-place modification, this needs adjustment.
    # For now, let's assume the caller will handle assignment.
    # If the C++ `curve = GeomConvert::CurveToBSplineCurve(c);` implies assignment,
    # then we should return the new curve.
    return new_curve


class CurvesToSurface:
    """
    Creates a surface by skinning a set of compatible B-spline curves.

    The curves must have compatible parameterizations, degrees, and
    knot vectors for successful skinning. This implementation aims to
    mirror the C++ version's functionality.
    """

    def __init__(
        self,
        curves: list[Geom_Curve],
        parameters: list[float] | None = None,
        continuous_if_closed: bool = False,
        tolerance: float = 1e-14,
    ):
        """
        Initialize the surface skinner.

        Args:
            curves: List of curves to be interpolated. These will be converted to B-splines.
            parameters: Optional list of parameters for v-direction interpolation. If None, they will be calculated.
            continuous_if_closed: If True, attempts to make a C2 continuous surface at the start/end junction
                                  if the first and last curve are the same.
            tolerance: Construction tolerance.
        """
        self._input_curves_raw = curves  # Store raw curves for potential conversion
        self._parameters = parameters if parameters is not None else []
        self._continuous_if_closed = continuous_if_closed
        self._tolerance = tolerance
        self._max_degree = (
            3  # Default max degree, matching C++ default for interpolation
        )

        self._has_performed = False
        self._skinned_surface = None

        self._input_curves: list[Geom_BSplineCurve] = []
        self._compatible_splines: list[Geom_BSplineCurve] = []

        # Convert all curves to bspline curves and store them
        for curve in self._input_curves_raw:
            self._input_curves.append(GeomConvert.CurveToBSplineCurve_s(curve))

        BSplineAlgorithms.match_degree(self._input_curves)

        # Calculate parameters if not provided
        if not self._parameters:
            self._calculate_parameters()

        # Ensure compatible splines for later use, if not already done by _calculate_parameters
        if not self._compatible_splines:  # Check if it's still empty
            self._compatible_splines = (
                BSplineAlgorithms.create_common_knots_vector_curve(
                    self._input_curves, self._tolerance
                )
            )

    def set_max_degree(self, degree: int):
        """
        Sets the maximum interpolation degree of the splines in skinning direction (v-direction).

        Args:
            degree: Maximum degree for v-direction interpolation.
        """
        if degree <= 0:
            raise ValueError("Degree must be positive.")
        self._max_degree = degree
        self.invalidate()  # Invalidate cached surface

    def get_parameters(self) -> list[float]:
        """Returns the parameters at the profile curves (v-direction)."""
        return self._parameters

    def surface(self) -> Geom_BSplineSurface:
        """Returns the skinned surface."""
        if not self._has_performed:
            self.perform()
        if self._skinned_surface is None:
            raise RuntimeError("Surface could not be created.")
        return self._skinned_surface

    def invalidate(self):
        """Invalidates the cached surface, forcing recomputation on next call."""
        self._has_performed = False
        self._skinned_surface = None

    def _calculate_parameters(self):
        """
        Calculates parameters for the v-direction based on control points.
        Mirrors C++ CurvesToSurface::CalculateParameters.
        """
        if not self._input_curves:
            return

        # Ensure common knot vector is created if not already
        if not self._compatible_splines:
            self._compatible_splines = (
                BSplineAlgorithms.create_common_knots_vector_curve(
                    self._input_curves, self._tolerance
                )
            )

        # Create a matrix of control points of all B-splines
        # (splines do have the same amount of control points now after matchDegree)
        first_curve = self._compatible_splines[0]
        num_poles_u = first_curve.NbPoles()
        num_splines = len(self._compatible_splines)

        # TColgp_Array2OfPnt(RowMin, RowMax, ColMin, ColMax)
        # Rows correspond to poles in U direction, Columns correspond to splines (V direction)
        control_points_matrix = TColgp_Array2OfPnt(1, num_poles_u, 1, num_splines)

        for spline_idx, spline in enumerate(self._compatible_splines, 1):
            for pole_idx in range(1, num_poles_u + 1):
                control_points_matrix.SetValue(
                    pole_idx, spline_idx, spline.Pole(pole_idx)
                )

        # Compute parameters using the surface-specific algorithm
        # C++ uses BSplineAlgorithms::computeParamsBSplineSurf(controlPoints)
        # Now that compute_params_bspline_surf is implemented in bspline_algorithms.py, use it.

        # The C++ version uses a default alpha of 0.5 for centripetal parameterization.
        # The `compute_params_bspline_surf` returns (u_params, v_params).
        # In CurvesToSurface, `_parameters` refers to the parameters in the v-direction.
        _, v_params = BSplineAlgorithms.compute_params_bspline_surf(
            control_points_matrix, alpha=0.5
        )
        self._parameters = v_params

    def perform(self):
        """
        Build the surface by skinning the curves.
        Mirrors C++ CurvesToSurface::Perform.
        """
        # check amount of given parameters
        if len(self._input_curves) < 2:
            return

        if len(self._parameters) != len(self._input_curves):
            raise error(
                "The amount of given parameters has to be equal to the amount of given B-splines!",
                ErrorCode.MATH_ERROR,
            )

        # check if all curves are closed
        # C++ uses BSplineAlgorithms::scale(_inputCurves) * BSplineAlgorithms::REL_TOL_CLOSED
        tolerance = (
            BSplineAlgorithms.scale(self._input_curves)
            * BSplineAlgorithms.REL_TOL_CLOSED
        )

        # Check if first and last input curves are equal
        make_closed = (
            self._continuous_if_closed
            and len(self._input_curves) > 0
            and self._input_curves[0].IsEqual(self._input_curves[-1], tolerance)
        )

        n_curves = len(self._input_curves)

        # create a common knot vector for all splines if not already done
        if not self._compatible_splines:
            self._compatible_splines = (
                BSplineAlgorithms.create_common_knots_vector_curve(
                    self._input_curves, self._tolerance
                )
            )

        first_curve = self._compatible_splines[0]
        num_control_points_u = first_curve.NbPoles()

        degree_v = 0
        degree_u = first_curve.Degree()
        knots_v = None
        mults_v = None

        # C++ uses Handle(TColgp_HArray2OfPnt) cpSurf;
        # Python equivalent: a list of lists or a numpy array, then convert to TColgp_Array2OfPnt
        # Or directly use TColgp_HArray1OfPnt for interpolation points.
        cp_surf = TColgp_Array2OfPnt()
        interp_spline = None

        # C++ uses Handle(TColgp_HArray1OfPnt) interpPointsVDir = new TColgp_HArray1OfPnt(1, static_cast<Standard_Integer>(nCurves));
        # This array will hold the poles for a given U index across all V curves.
        interp_points_v_dir = TColgp_HArray1OfPnt(1, n_curves)

        # C++ iterates from cpUIdx = 1 to numControlPointsU
        for cp_u_idx in range(1, num_control_points_u + 1):
            # Populate interpPointsVDir with poles from each compatible spline at the current U index
            for cp_v_idx in range(1, n_curves + 1):
                pole = self._compatible_splines[cp_v_idx - 1].Pole(cp_u_idx)
                interp_points_v_dir.SetValue(cp_v_idx, pole)

            # C++ uses PointsToBSplineInterpolation interpol(interpPointsVDir, _parameters, _maxDegree, makeClosed);
            # Python equivalent: PointsToBSplineInterpolation(points, parameters, max_degree, continuous_if_closed)
            interpolator = PointsToBSplineInterpolation(
                points=interp_points_v_dir,
                parameters=self._parameters,
                max_degree=self._max_degree,
                continuous_if_closed=make_closed,
            )

            # Call curve() to get the interpolated spline
            interp_spline = interpolator.curve()

            if interp_spline is None:
                raise RuntimeError(f"Interpolation failed for U index {cp_u_idx}")

            if make_closed:
                # Apply clamping if the curve is closed and we need continuity
                clamped_spline = clamp_bspline(interp_spline)
                if (
                    clamped_spline
                ):  # clamp_bspline returns None if not periodic or if modification failed
                    interp_spline = clamped_spline

            if cp_u_idx == 1:
                degree_v = interp_spline.Degree()

                # Get knots and multiplicities for V direction
                knots_v_list = []
                for i in range(1, interp_spline.NbKnots() + 1):
                    knots_v_list.append(interp_spline.Knot(i))
                knots_v = TColStd_Array1OfReal(1, len(knots_v_list))
                for i, k in enumerate(knots_v_list, 1):
                    knots_v.SetValue(i, k)

                mults_v_list = []
                for i in range(1, interp_spline.NbKnots() + 1):
                    mults_v_list.append(interp_spline.Multiplicity(i))
                mults_v = TColStd_Array1OfInteger(1, len(mults_v_list))
                for i, m in enumerate(mults_v_list, 1):
                    mults_v.SetValue(i, m)

                # Initialize the surface control points array
                # C++ uses Handle(TColgp_HArray2OfPnt) cpSurf;
                # cpSurf = new TColgp_HArray2OfPnt(1, static_cast<Standard_Integer>(numControlPointsU), 1, interpSpline->NbPoles());
                # Python equivalent: TColgp_Array2OfPnt(RowMin, RowMax, ColMin, ColMax)
                # Rows = U direction (num_control_points_u), Cols = V direction (interpSpline.NbPoles())
                cp_surf = TColgp_Array2OfPnt(
                    1, num_control_points_u, 1, interp_spline.NbPoles()
                )

            # the final surface control points are the control points resulting from the interpolation
            for i in range(1, interp_spline.NbPoles() + 1):
                cp_surf.SetValue(cp_u_idx, i, interp_spline.Pole(i))

            # check degree always the same
            # C++ asserts degreeV == interpSpline->Degree()
            if cp_u_idx > 1 and degree_v != interp_spline.Degree():
                raise error(
                    f"Inconsistent degree_v found at U index {cp_u_idx}. Expected {degree_v}, got {interp_spline.Degree()}.",
                    ErrorCode.MATH_ERROR,
                )

        # Get U knots and multiplicities from the first compatible spline
        knots_u_list = []
        for i in range(1, first_curve.NbKnots() + 1):
            knots_u_list.append(first_curve.Knot(i))
        knots_u = TColStd_Array1OfReal(1, len(knots_u_list))
        for i, k in enumerate(knots_u_list, 1):
            knots_u.SetValue(i, k)

        mults_u_list = []
        for i in range(1, first_curve.NbKnots() + 1):
            mults_u_list.append(first_curve.Multiplicity(i))
        mults_u = TColStd_Array1OfInteger(1, len(mults_u_list))
        for i, m in enumerate(mults_u_list, 1):
            mults_u.SetValue(i, m)

        # Construct the final Geom_BSplineSurface

        # Check if knots_v and mults_v were successfully populated
        if knots_v is None or mults_v is None:
            raise RuntimeError("Failed to obtain V direction knots and multiplicities.")

        # Number of poles in U direction: num_control_points_u
        # Number of poles in V direction: interp_spline.NbPoles()

        self._skinned_surface = Geom_BSplineSurface(
            cp_surf,
            knots_u,
            knots_v,
            mults_u,
            mults_v,
            degree_u,
            degree_v,
        )

        self._has_performed = True



import math
import os
import sys

import numpy as np
import pytest
from OCP.Geom import (
    Geom_BSplineCurve,
    Geom_BSplineSurface,
    Geom_Circle,
    Geom_Curve,
    Geom_Ellipse,
    Geom_TrimmedCurve,
)
from OCP.GeomAPI import GeomAPI_ProjectPointOnCurve
from OCP.gp import gp_Ax2, gp_Dir, gp_Pnt
from OCP.Precision import Precision
from OCP.TColgp import TColgp_Array1OfPnt, TColgp_Array2OfPnt, TColgp_HArray1OfPnt
from OCP.TColStd import TColStd_Array1OfInteger, TColStd_Array1OfReal

# Add the parent directory to the path to import the module
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

from src_py.ocp_gordon.internal.bspline_algorithms import (
    BSplineAlgorithms,
    SurfaceDirection,
)
from src_py.ocp_gordon.internal.misc import clone_bspline


# Helper for comparing lists of floats
def assert_list_almost_equal(list1, list2, places=7):
    assert len(list1) == len(list2)
    for i in range(len(list1)):
        assert math.isclose(list1[i], list2[i], rel_tol=10 ** (-places))


# sum of mults = num_control_points + degree + 1
# first and last values of mults = degree + 1
@pytest.fixture
def simple_bspline_curve():
    poles = TColgp_Array1OfPnt(1, 4)
    poles.SetValue(1, gp_Pnt(0, 0, 0))
    poles.SetValue(2, gp_Pnt(1, 1, 0))
    poles.SetValue(3, gp_Pnt(2, -1, 0))
    poles.SetValue(4, gp_Pnt(3, 0, 0))

    knots = TColStd_Array1OfReal(1, 3)
    knots.SetValue(1, 0.0)
    knots.SetValue(2, 0.5)
    knots.SetValue(3, 1.0)

    mults = TColStd_Array1OfInteger(1, 3)
    mults.SetValue(1, 3)
    mults.SetValue(2, 1)
    mults.SetValue(3, 3)

    return Geom_BSplineCurve(poles, knots, mults, 2)  # Degree 2


def test_linspace_with_breaks():
    umin, umax = 0.0, 1.0
    n_values = 5
    breaks = [0.25, 0.75]
    result = BSplineAlgorithms.linspace_with_breaks(umin, umax, n_values, breaks)

    assert math.isclose(result[0], 0.0)
    assert math.isclose(result[1], 0.25)
    assert math.isclose(result[2], 0.5)
    assert math.isclose(result[3], 0.75)
    assert math.isclose(result[4], 1.0)
    assert len(result) == 5

    breaks_existing = [0.0, 0.5, 1.0]
    result_existing = BSplineAlgorithms.linspace_with_breaks(
        umin, umax, n_values, breaks_existing
    )
    assert len(result_existing) == 5
    assert_list_almost_equal(result_existing, [0.0, 0.25, 0.5, 0.75, 1.0])

    breaks_outside = [-0.1, 1.1]
    result_outside = BSplineAlgorithms.linspace_with_breaks(
        umin, umax, n_values, breaks_outside
    )
    assert math.isclose(result_outside[0], -0.1)
    assert math.isclose(result_outside[1], 0.0)
    assert math.isclose(result_outside[-1], 1.1)
    assert len(result_outside) == 7


def test_is_u_dir_closed():
    poles_closed_u = TColgp_Array2OfPnt(1, 2, 1, 2)
    poles_closed_u.SetValue(1, 1, gp_Pnt(0, 0, 0))
    poles_closed_u.SetValue(2, 1, gp_Pnt(0, 1, 0))
    poles_closed_u.SetValue(1, 2, gp_Pnt(1, 0, 0))
    poles_closed_u.SetValue(2, 2, gp_Pnt(1, 1, 0))

    poles_closed_u.SetValue(2, 1, gp_Pnt(0, 0, 0))
    poles_closed_u.SetValue(2, 2, gp_Pnt(1, 0, 0))

    assert BSplineAlgorithms.is_u_dir_closed(poles_closed_u, 1e-6)

    poles_not_closed_u = TColgp_Array2OfPnt(1, 2, 1, 2)
    poles_not_closed_u.SetValue(1, 1, gp_Pnt(0, 0, 0))
    poles_not_closed_u.SetValue(2, 1, gp_Pnt(0, 1, 0))
    poles_not_closed_u.SetValue(1, 2, gp_Pnt(1, 0, 0))
    poles_not_closed_u.SetValue(2, 2, gp_Pnt(1, 1, 0))
    assert not BSplineAlgorithms.is_u_dir_closed(poles_not_closed_u, 1e-6)


def test_is_v_dir_closed():
    poles_closed_v = TColgp_Array2OfPnt(1, 2, 1, 2)
    poles_closed_v.SetValue(1, 1, gp_Pnt(0, 0, 0))
    poles_closed_v.SetValue(1, 2, gp_Pnt(1, 0, 0))
    poles_closed_v.SetValue(2, 1, gp_Pnt(0, 1, 0))
    poles_closed_v.SetValue(2, 2, gp_Pnt(1, 1, 0))

    poles_closed_v.SetValue(1, 2, gp_Pnt(0, 0, 0))
    poles_closed_v.SetValue(2, 2, gp_Pnt(0, 1, 0))

    assert BSplineAlgorithms.is_v_dir_closed(poles_closed_v, 1e-6)

    poles_not_closed_v = TColgp_Array2OfPnt(1, 2, 1, 2)
    poles_not_closed_v.SetValue(1, 1, gp_Pnt(0, 0, 0))
    poles_not_closed_v.SetValue(1, 2, gp_Pnt(1, 0, 0))
    poles_not_closed_v.SetValue(2, 1, gp_Pnt(0, 1, 0))
    poles_not_closed_v.SetValue(2, 2, gp_Pnt(1, 1, 0))
    assert not BSplineAlgorithms.is_v_dir_closed(poles_not_closed_v, 1e-6)


def test_compute_params_bspline_curve():
    points_h_array = TColgp_HArray1OfPnt(1, 3)
    points_h_array.SetValue(1, gp_Pnt(0, 0, 0))
    points_h_array.SetValue(2, gp_Pnt(1, 0, 0))
    points_h_array.SetValue(3, gp_Pnt(2, 0, 0))

    params = BSplineAlgorithms.compute_params_bspline_curve(points_h_array, alpha=0.5)
    assert len(params) == 3
    assert math.isclose(params[0], 0.0)
    assert math.isclose(params[1], 0.5)
    assert math.isclose(params[2], 1.0)

    points_h_array_non_uniform = TColgp_HArray1OfPnt(1, 4)
    points_h_array_non_uniform.SetValue(1, gp_Pnt(0, 0, 0))
    points_h_array_non_uniform.SetValue(2, gp_Pnt(1, 0, 0))
    points_h_array_non_uniform.SetValue(3, gp_Pnt(3, 0, 0))
    points_h_array_non_uniform.SetValue(4, gp_Pnt(4, 0, 0))

    params_non_uniform = BSplineAlgorithms.compute_params_bspline_curve(
        points_h_array_non_uniform, alpha=0.5
    )
    assert len(params_non_uniform) == 4
    assert math.isclose(params_non_uniform[0], 0.0)
    assert math.isclose(params_non_uniform[1], 1.0 / (1.0 + math.sqrt(2.0) + 1.0))
    assert math.isclose(
        params_non_uniform[2], (1.0 + math.sqrt(2.0)) / (1.0 + math.sqrt(2.0) + 1.0)
    )
    assert math.isclose(params_non_uniform[3], 1.0)


def test_get_kink_parameters(simple_bspline_curve):
    kinks = BSplineAlgorithms.get_kink_parameters(simple_bspline_curve)
    assert len(kinks) == 0

    poles_kink = TColgp_Array1OfPnt(1, 5)
    poles_kink.SetValue(1, gp_Pnt(0, 0, 0))
    poles_kink.SetValue(2, gp_Pnt(1, 1, 0))
    poles_kink.SetValue(3, gp_Pnt(1, 0, 0))
    poles_kink.SetValue(4, gp_Pnt(2, 1, 0))
    poles_kink.SetValue(5, gp_Pnt(3, 0, 0))

    knots_kink = TColStd_Array1OfReal(1, 3)
    knots_kink.SetValue(1, 0.0)
    knots_kink.SetValue(2, 0.5)
    knots_kink.SetValue(3, 1.0)

    mults_kink = TColStd_Array1OfInteger(1, 3)
    mults_kink.SetValue(1, 3)
    mults_kink.SetValue(2, 2)
    mults_kink.SetValue(3, 3)

    curve_with_kink = Geom_BSplineCurve(poles_kink, knots_kink, mults_kink, 2)

    kinks_found = BSplineAlgorithms.get_kink_parameters(curve_with_kink)
    assert len(kinks_found) == 1
    assert math.isclose(kinks_found[0], 0.5)


def test_trim_curve(simple_bspline_curve):
    trimmed_curve = BSplineAlgorithms.trim_curve(simple_bspline_curve, 0.25, 0.75)
    assert trimmed_curve is not None
    assert math.isclose(trimmed_curve.FirstParameter(), 0.25)
    assert math.isclose(trimmed_curve.LastParameter(), 0.75)
    assert not (trimmed_curve == simple_bspline_curve)


def test_match_parameter_range(simple_bspline_curve):
    curve1 = clone_bspline(simple_bspline_curve)
    curve2_poles = TColgp_Array1OfPnt(1, 4)
    curve2_poles.SetValue(1, gp_Pnt(0, 0, 0))
    curve2_poles.SetValue(2, gp_Pnt(1, 1, 0))
    curve2_poles.SetValue(3, gp_Pnt(2, -1, 0))
    curve2_poles.SetValue(4, gp_Pnt(3, 0, 0))
    curve2_knots = TColStd_Array1OfReal(1, 3)
    curve2_knots.SetValue(1, 0.5)
    curve2_knots.SetValue(2, 1.0)
    curve2_knots.SetValue(3, 1.5)
    curve2_mults = TColStd_Array1OfInteger(1, 3)
    curve2_mults.SetValue(1, 3)
    curve2_mults.SetValue(2, 1)
    curve2_mults.SetValue(3, 3)
    curve2 = Geom_BSplineCurve(curve2_poles, curve2_knots, curve2_mults, 2)

    curves = [curve1, curve2]
    BSplineAlgorithms.match_parameter_range(curves)

    assert math.isclose(curves[0].FirstParameter(), curves[1].FirstParameter())
    assert math.isclose(curves[0].LastParameter(), curves[1].LastParameter())
    assert math.isclose(
        curves[0].FirstParameter(), simple_bspline_curve.FirstParameter()
    )
    assert math.isclose(curves[0].LastParameter(), simple_bspline_curve.LastParameter())


def test_match_degree(simple_bspline_curve):
    curve1 = clone_bspline(simple_bspline_curve)

    poles_deg3 = TColgp_Array1OfPnt(1, 5)
    poles_deg3.SetValue(1, gp_Pnt(0, 0, 0))
    poles_deg3.SetValue(2, gp_Pnt(1, 1, 0))
    poles_deg3.SetValue(3, gp_Pnt(2, -1, 0))
    poles_deg3.SetValue(4, gp_Pnt(3, 1, 0))
    poles_deg3.SetValue(5, gp_Pnt(4, 0, 0))
    knots_deg3 = TColStd_Array1OfReal(1, 3)
    knots_deg3.SetValue(1, 0.0)
    knots_deg3.SetValue(2, 0.5)
    knots_deg3.SetValue(3, 1.0)
    mults_deg3 = TColStd_Array1OfInteger(1, 3)
    mults_deg3.SetValue(1, 4)
    mults_deg3.SetValue(2, 1)
    mults_deg3.SetValue(3, 4)
    curve2 = Geom_BSplineCurve(poles_deg3, knots_deg3, mults_deg3, 3)

    curves = [curve1, curve2]
    BSplineAlgorithms.match_degree(curves)

    assert curves[0].Degree() == 3
    assert curves[1].Degree() == 3


def test_reparametrize_bspline(simple_bspline_curve):
    curve = clone_bspline(simple_bspline_curve)
    BSplineAlgorithms.reparametrize_bspline(curve, 10.0, 20.0)
    assert math.isclose(curve.FirstParameter(), 10.0)
    assert math.isclose(curve.LastParameter(), 20.0)

    original_point_at_0_5 = simple_bspline_curve.Value(0.5)
    reparam_point_at_15_0 = curve.Value(15.0)
    assert original_point_at_0_5.IsEqual(reparam_point_at_15_0, 1e-6)


def test_convert_to_bspline():
    def distance_curve_to_curve(c1: Geom_Curve, c2: Geom_Curve):
        min_distance = 0.0
        n_samples = 500
        for i in range(n_samples):
            u = (1 - i / n_samples) * c1.FirstParameter() + (
                i / n_samples
            ) * c1.LastParameter()
            pnt = c1.Value(u)
            projection = GeomAPI_ProjectPointOnCurve(pnt, c2)
            min_distance = max(min_distance, projection.LowerDistance())
        return min_distance

    def check_trimmed_curve(curve: Geom_Curve, u1: float, u2: float, curve_size: float):

        curve = Geom_TrimmedCurve(curve, u1, u2)
        bspline = BSplineAlgorithms._convert_to_bspline(curve)
        assert (
            distance_curve_to_curve(curve, bspline)
            < Precision.Approximation_s() * curve_size / 200
        )

    def check_circle(radius: float):
        curve = Geom_Circle(gp_Ax2(gp_Pnt(0, 0, 0), gp_Dir(0, -1, 1)), radius)
        bspline = BSplineAlgorithms._convert_to_bspline(curve)
        assert (
            distance_curve_to_curve(curve, bspline)
            < Precision.Approximation_s() * 2 * radius / 200
        )
        u1, u2 = (
            curve.FirstParameter(),
            0.5 * curve.FirstParameter() + 0.5 * curve.LastParameter(),
        )
        check_trimmed_curve(curve, u1, u2, 2 * radius)

    check_circle(10)
    check_circle(100)
    check_circle(1000)

    def check_ellipse(major_radius: float, minor_radius: float):
        curve = Geom_Ellipse(
            gp_Ax2(gp_Pnt(0, 0, 0), gp_Dir(0, 1, 1)), major_radius, minor_radius
        )
        bspline = BSplineAlgorithms._convert_to_bspline(curve)
        assert (
            distance_curve_to_curve(curve, bspline)
            < Precision.Approximation_s() * 2 * major_radius / 200
        )
        u1, u2 = (
            curve.FirstParameter(),
            0.5 * curve.FirstParameter() + 0.5 * curve.LastParameter(),
        )
        check_trimmed_curve(curve, u1, u2, 2 * major_radius)

    check_ellipse(10, 5)
    check_ellipse(100, 25)
    check_ellipse(1000, 750)


def test_to_bsplines(simple_bspline_curve):
    curves = [simple_bspline_curve]
    bsplines = BSplineAlgorithms.to_bsplines(curves)
    assert len(bsplines) == 1
    assert isinstance(bsplines[0], Geom_BSplineCurve)


def test_scale_single_curve(simple_bspline_curve):
    scale = BSplineAlgorithms.scale(simple_bspline_curve)
    assert scale > 0.0
    assert scale < 10.0


def test_scale_list_of_curves(simple_bspline_curve):
    curve1 = clone_bspline(simple_bspline_curve)
    curve2_poles = TColgp_Array1OfPnt(1, 2)
    curve2_poles.SetValue(1, gp_Pnt(0, 0, 0))
    curve2_poles.SetValue(2, gp_Pnt(10, 0, 0))
    curve2_knots = TColStd_Array1OfReal(1, 2)
    curve2_knots.SetValue(1, 0.0)
    curve2_knots.SetValue(2, 1.0)
    curve2_mults = TColStd_Array1OfInteger(1, 2)
    curve2_mults.SetValue(1, 2)
    curve2_mults.SetValue(2, 2)
    curve2 = Geom_BSplineCurve(curve2_poles, curve2_knots, curve2_mults, 1)

    curves = [curve1, curve2]
    scale = BSplineAlgorithms.scale(curves)
    assert scale > 0.0
    assert math.isclose(scale, 10.0, rel_tol=0.1)


def test_get_kink_parameters_surface():
    poles_surf = TColgp_Array2OfPnt(1, 2, 1, 2)
    poles_surf.SetValue(1, 1, gp_Pnt(0, 0, 0))
    poles_surf.SetValue(1, 2, gp_Pnt(1, 0, 0))
    poles_surf.SetValue(2, 1, gp_Pnt(0, 1, 0))
    poles_surf.SetValue(2, 2, gp_Pnt(1, 1, 0))

    knots_u = TColStd_Array1OfReal(1, 2)
    knots_u.SetValue(1, 0.0)
    knots_u.SetValue(2, 1.0)
    mults_u = TColStd_Array1OfInteger(1, 2)
    mults_u.SetValue(1, 2)
    mults_u.SetValue(2, 2)

    knots_v = TColStd_Array1OfReal(1, 2)
    knots_v.SetValue(1, 0.0)
    knots_v.SetValue(2, 1.0)
    mults_v = TColStd_Array1OfInteger(1, 2)
    mults_v.SetValue(1, 2)
    mults_v.SetValue(2, 2)

    test_surface = Geom_BSplineSurface(
        poles_surf, knots_u, knots_v, mults_u, mults_v, 1, 1
    )

    kinks = BSplineAlgorithms.get_kink_parameters_surface(test_surface)
    assert len(kinks.u) == 0
    assert len(kinks.v) == 0


def test_knots_from_curve_parameters_open():
    params = [0.0, 0.25, 0.5, 0.75, 1.0]
    degree = 2
    knots = BSplineAlgorithms.knots_from_curve_parameters(
        params, degree, closed_curve=False
    )
    assert len(knots) == len(params) + degree + 1
    assert_list_almost_equal(knots, [0.0, 0.0, 0.0, 0.375, 0.625, 1.0, 1.0, 1.0])


def test_knots_from_curve_parameters_closed_even_degree():
    params = [0.0, 0.25, 0.5, 0.75, 1.0]
    degree = 2
    params_copy = list(params)
    knots = BSplineAlgorithms.knots_from_curve_parameters(
        params_copy, degree, closed_curve=True
    )
    assert len(knots) == 9
    assert knots[0] <= knots[1]
    assert knots[-1] >= knots[-2]


def test_knots_from_curve_parameters_closed_odd_degree():
    params = [0.0, 0.25, 0.5, 0.75, 1.0]
    degree = 3
    params_copy = list(params)
    knots = BSplineAlgorithms.knots_from_curve_parameters(
        params_copy, degree, closed_curve=True
    )
    assert len(knots) == 11
    assert knots[0] <= knots[1]
    assert knots[-1] >= knots[-2]


def test_compute_params_bspline_surf():
    poles_surf = TColgp_Array2OfPnt(1, 3, 1, 3)
    poles_surf.SetValue(1, 1, gp_Pnt(0, 0, 0))
    poles_surf.SetValue(1, 2, gp_Pnt(1, 0, 0))
    poles_surf.SetValue(1, 3, gp_Pnt(2, 0, 0))
    poles_surf.SetValue(2, 1, gp_Pnt(0, 1, 0))
    poles_surf.SetValue(2, 2, gp_Pnt(1, 1, 0))
    poles_surf.SetValue(2, 3, gp_Pnt(2, 1, 0))
    poles_surf.SetValue(3, 1, gp_Pnt(0, 2, 0))
    poles_surf.SetValue(3, 2, gp_Pnt(1, 2, 0))
    poles_surf.SetValue(3, 3, gp_Pnt(2, 2, 0))

    params_u, params_v = BSplineAlgorithms.compute_params_bspline_surf(
        poles_surf, alpha=0.5
    )

    assert len(params_u) == 3
    assert len(params_v) == 3
    assert_list_almost_equal(params_u, [0.0, 0.5, 1.0])
    assert_list_almost_equal(params_v, [0.0, 0.5, 1.0])


if __name__ == "__main__":
    # pytest.main([f'{__file__}::test_trim_curve', "-v"])
    pytest.main([f"{__file__}", "-v"])



"""
Point interpolation to B-spline curves.

This module provides functionality to create B-spline curves
that interpolate a set of points using the Park (2000) algorithm
for node and knot selection.
"""

import numpy as np
from OCP.Geom import Geom_BSplineCurve, Geom_TrimmedCurve
from OCP.GeomConvert import GeomConvert
from OCP.BSplCLib import BSplCLib
from OCP.TColgp import TColgp_Array1OfPnt, TColgp_HArray1OfPnt
from OCP.TColStd import TColStd_Array1OfReal, TColStd_Array1OfInteger, TColStd_HArray1OfReal
from OCP.gp import gp_Pnt
from typing import List, Tuple, Optional, Union

from .bspline_algorithms import BSplineAlgorithms


class PointsToBSplineInterpolation:
    """
    Creates a B-spline curve that interpolates a set of points.
    
    Implements the b-spline interpolation algorithm as described by
    Park (2000): Choosing nodes and knots in closed B-spline curve interpolation to point data
    """
    
    def __init__(self,
                 points: TColgp_HArray1OfPnt,
                 parameters: list[float] | None = None,
                 max_degree: int = 3,
                 continuous_if_closed: bool = False):
        """
        Initialize the point interpolator.
        
        Args:
            points: Points to interpolate
            parameters: Custom parameter values (optional)
            max_degree: Maximum degree of the B-spline
            continuous_if_closed: Whether to ensure C2 continuity for closed curves
        """
        self.m_pnts = points
        self.m_degree = max_degree
        self.m_C2Continuous = continuous_if_closed
        
        # Use provided parameters or compute them
        if parameters is not None:
            self.m_params = parameters
        else:
            self.m_params = BSplineAlgorithms.compute_params_bspline_curve(points)
        
        # Validate inputs
        self._validate_parameters()
    
    def _validate_parameters(self):
        """Validate input parameters."""
        if self.m_degree < 1:
            raise ValueError("Degree must be larger than 1 in PointsToBSplineInterpolation!")
        
        if self.m_pnts is None:
            raise ValueError("No points given in PointsToBSplineInterpolation")
        
        if self.m_pnts.Length() < 2:
            raise ValueError("Too few points in PointsToBSplineInterpolation")
        
        if len(self.m_params) != self.m_pnts.Length():
            raise ValueError(f"Number of parameters {len(self.m_params)} and points {self.m_pnts.Length()} don't match in PointsToBSplineInterpolation")
    
    def curve(self) -> Geom_BSplineCurve:
        """Returns the interpolation curve using Park (2000) algorithm."""
        degree = self.degree()
        
        # Work with a copy of parameters
        params = self.m_params.copy()
        
        # Generate knots using Park (2000) algorithm
        knots = BSplineAlgorithms.knots_from_curve_parameters(
            params, degree, self.is_closed()
        )
        
        if self.is_closed():
            # Remove the last parameter for closed curves (implicitly included by wrapping)
            params = params[:-1]
        
        # Convert to OpenCASCADE arrays
        occ_knots = BSplineAlgorithms.to_array(knots)
        occ_params = BSplineAlgorithms.to_array(params)
        
        # Compute B-spline basis matrix
        # bspl_mat is a numpy array (n_params x n_params + degree)
        bspl_mat = BSplineAlgorithms.bspline_basis_mat(
            degree, occ_knots.Array1(), occ_params.Array1()
        )
        
        # Build left hand side of the linear system (A * C = P)
        # A is the basis matrix, C are control points, P are interpolation points
        n_params = len(params)
        lhs = np.zeros((n_params, n_params))
        
        # Fill the matrix with basis function values
        for i in range(n_params):
            for j in range(n_params):
                lhs[i, j] = bspl_mat[i, j]
        
        if self.is_closed():
            # Set continuity constraints for closed curves by wrapping control points
            # The first 'degree' control points are influenced by the last 'degree' basis functions
            # and vice-versa. This effectively wraps the control points.
            for j in range(degree):
                for i in range(n_params):
                    lhs[i, j] += bspl_mat[i, n_params + j]
        
        # Right hand side - point coordinates
        # Store points in a numpy array for easier manipulation
        rhs_pnts = np.zeros((n_params, 3)) # (x, y, z)
        
        for i in range(n_params):
            p = self.m_pnts(i + 1) # OCP arrays are 1-indexed
            rhs_pnts[i, 0] = p.X()
            rhs_pnts[i, 1] = p.Y()
            rhs_pnts[i, 2] = p.Z()
        
        # Solve linear system using numpy.linalg.solve
        try:
            # cp_coords will be (n_params, 3) where each row is (cx, cy, cz)
            cp_coords = np.linalg.solve(lhs, rhs_pnts)
        except np.linalg.LinAlgError as e:
            raise ValueError(f"Singular Matrix in linear system solution: {e}")
        
        # Create control points
        if self.is_closed():
            n_ctrl_pnts = len(self.m_params) + degree - 1
        else:
            n_ctrl_pnts = n_params
        
        if self.needs_shifting():
            n_ctrl_pnts += 1
        
        poles = TColgp_Array1OfPnt(1, n_ctrl_pnts)
        
        # Fill control points from solution
        for i in range(n_params):
            pnt = gp_Pnt(cp_coords[i, 0], cp_coords[i, 1], cp_coords[i, 2])
            poles.SetValue(i + 1, pnt)
        
        if self.is_closed():
            # Wrap control points for closed curves
            # The first 'degree' control points are duplicates of the last 'degree'
            # from the solution, to ensure C2 continuity.
            for i in range(degree): # Only degree points are wrapped
                pnt = gp_Pnt(cp_coords[i, 0], cp_coords[i, 1], cp_coords[i, 2])
                poles.SetValue(n_params + i + 1, pnt)
        
        if self.needs_shifting():
            # Add extra control point and knot for even-degree closed curves
            deg = degree
            knots.append(knots[-1] + knots[2 * deg + 1] - knots[2 * deg])
            poles.SetValue(n_params + degree + 1, poles(degree + 1))
            
            # Shift knots back
            for i in range(len(knots)):
                knots[i] -= params[0]
        
        # Extract unique knots and their multiplicities from the flat knot vector

        occ_knots = BSplineAlgorithms.to_array(knots)
        knotsLen = BSplCLib.KnotsLength_s(occ_knots)

        occ_unique_knots = TColStd_Array1OfReal(1, knotsLen)
        occ_multiplicities = TColStd_Array1OfInteger(1, knotsLen)
        BSplCLib.Knots_s(occ_knots, occ_unique_knots, occ_multiplicities)

        # print(f'params={params}')
        # print(f'knots={knots}')
        # print(f'n_ctrl_pnts={n_ctrl_pnts}, unique_knots={occ_unique_knots}, multiplicities={occ_multiplicities}, degree={degree}')
        
        # Create final B-spline curve
        result = Geom_BSplineCurve(
            poles, occ_unique_knots, occ_multiplicities, degree, False
        )
        
        # clamp bspline
        if self.is_closed():
            result = PointsToBSplineInterpolation.clamp(result, self.m_params[0], self.m_params[-1])
        
        return result
    
    def degree(self) -> int:
        """Return the degree of the B-spline interpolation."""
        max_degree = self.m_pnts.Length() - 1
        if self.is_closed():
            max_degree -= 1
        degree = min(max_degree, self.m_degree)
        if degree <= 0:
            raise ValueError("Invalid degree computed")
        return degree
    
    def is_closed(self) -> bool:
        """Check if the curve is closed."""
        return BSplineAlgorithms.is_closed(self.m_pnts, self.m_C2Continuous)
    
    def needs_shifting(self) -> bool:
        """Check if knot shifting is needed for closed curves."""
        return (self.degree() % 2) == 0 and self.is_closed()
    
    def __call__(self) -> Geom_BSplineCurve:
        """Allow calling the object directly to get the curve."""
        return self.curve()

    @staticmethod
    def clamp(curve: Geom_BSplineCurve, first_param: float, last_param: float, tol: float = 1e-7) -> Geom_BSplineCurve:
        c = Geom_TrimmedCurve(curve, first_param, last_param)
        curve = GeomConvert.CurveToBSplineCurve_s(c)
        return curve



import pytest
import sys
import os
import numpy as np
from OCP.Geom import Geom_BSplineCurve, Geom_Curve
from OCP.TColgp import TColgp_Array1OfPnt
from OCP.gp import gp_Pnt
import math

# Add the parent directory to the path to import the module
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '..'))

from src_py.ocp_gordon.internal.bspline_approx_interp import BSplineApproxInterp, ProjectResult
from src_py.ocp_gordon.internal.error import error, ErrorCode
from src_py.ocp_gordon.internal.bspline_algorithms import BSplineAlgorithms

# Helper for comparing lists of floats
def assert_list_almost_equal(list1, list2, places=7):
    assert len(list1) == len(list2)
    for i in range(len(list1)):
        assert math.isclose(list1[i], list2[i], rel_tol=10**(-places))

@pytest.fixture
def simple_points_array():
    points = TColgp_Array1OfPnt(1, 4)
    points.SetValue(1, gp_Pnt(0, 0, 0))
    points.SetValue(2, gp_Pnt(1, 1, 0))
    points.SetValue(3, gp_Pnt(2, -1, 0))
    points.SetValue(4, gp_Pnt(3, 0, 0))
    return points

@pytest.fixture
def linear_points_array():
    points = TColgp_Array1OfPnt(1, 3)
    points.SetValue(1, gp_Pnt(0, 0, 0))
    points.SetValue(2, gp_Pnt(1, 0, 0))
    points.SetValue(3, gp_Pnt(2, 0, 0))
    return points

@pytest.fixture
def closed_points_array():
    points = TColgp_Array1OfPnt(1, 5)
    points.SetValue(1, gp_Pnt(0, 0, 0))
    points.SetValue(2, gp_Pnt(1, 0, 0))
    points.SetValue(3, gp_Pnt(1, 1, 0))
    points.SetValue(4, gp_Pnt(0, 1, 0))
    points.SetValue(5, gp_Pnt(0, 0, 0)) # Closed loop
    return points

class TestBSplineApproxInterp:

    def test_init(self, simple_points_array):
        interp = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=3)
        assert interp.m_pnts.Length() == 4
        assert len(interp.m_index_of_approximated) == 4
        assert len(interp.m_index_of_interpolated) == 0
        assert interp.m_degree == 3
        assert interp.m_ncp == 5
        assert not interp.m_c2_continuous

    def test_interpolate_point(self, simple_points_array):
        interp = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=3)
        interp.interpolate_point(1)
        assert 1 not in interp.m_index_of_approximated
        assert 1 in interp.m_index_of_interpolated
        assert len(interp.m_index_of_kinks) == 0

        interp.interpolate_point(2, with_kink=True)
        assert 2 not in interp.m_index_of_approximated
        assert 2 in interp.m_index_of_interpolated
        assert 2 in interp.m_index_of_kinks

        with pytest.raises(error) as excinfo:
            interp.interpolate_point(1) # Already interpolated
        assert excinfo.value.get_code() == ErrorCode.INDEX_ERROR

    def test_max_distance_of_bounding_box(self, simple_points_array):
        interp = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=3)
        dist = interp._max_distance_of_bounding_box(simple_points_array)
        # Bounding box from (0,-1,0) to (3,1,0)
        expected_dist = gp_Pnt(0,-1,0).Distance(gp_Pnt(3,1,0))
        assert math.isclose(dist, expected_dist, rel_tol=1e-6)

    def test_is_closed(self, simple_points_array, closed_points_array):
        # Not C2 continuous, so should return False even if points are closed
        interp_not_c2 = BSplineApproxInterp(closed_points_array, n_control_points=5, degree=3, continuous_if_closed=False)
        assert not interp_not_c2.is_closed()

        # C2 continuous, and points are closed
        interp_c2 = BSplineApproxInterp(closed_points_array, n_control_points=5, degree=3, continuous_if_closed=True)
        assert interp_c2.is_closed()

        # Not closed points
        interp_open = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=3, continuous_if_closed=True)
        assert not interp_open.is_closed()

    def test_first_and_last_interpolated(self, simple_points_array):
        interp = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=3)
        assert not interp._first_and_last_interpolated()

        interp.interpolate_point(0)
        assert not interp._first_and_last_interpolated()

        interp.interpolate_point(simple_points_array.Length() - 1)
        assert interp._first_and_last_interpolated()

    def test_compute_parameters(self, simple_points_array, linear_points_array):
        interp = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=3)
        params = interp._compute_parameters(0.5)
        assert len(params) == 4
        assert math.isclose(params[0], 0.0)
        assert math.isclose(params[-1], 1.0)
        # Specific values depend on point distances, check for monotonicity
        assert params[0] <= params[1] <= params[2] <= params[3]

        interp_linear = BSplineApproxInterp(linear_points_array, n_control_points=3, degree=1)
        params_linear = interp_linear._compute_parameters(0.5)
        assert_list_almost_equal(params_linear, [0.0, 0.5, 1.0])

    def test_compute_knots(self, simple_points_array):
        interp = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=3)
        params = interp._compute_parameters(0.5)
        knots, mults = interp._compute_knots(interp.m_ncp, params)

        # For a clamped B-spline: sum of mults = ncp + degree + 1
        # ncp = 5, degree = 3. Sum of mults should be 5 + 3 + 1 = 9
        assert sum(mults) == interp.m_ncp + interp.m_degree + 1
        # First and last mult should be degree + 1 = 4
        assert mults[0] == interp.m_degree + 1
        assert mults[-1] == interp.m_degree + 1
        # Number of distinct knots = ncp - degree + 1 = 5 - 3 + 1 = 3
        assert len(knots) == interp.m_ncp - interp.m_degree + 1

        # Test with kinks
        interp_kink = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=3)
        interp_kink.interpolate_point(1, with_kink=True) # Add a kink at index 1
        knots_kink, mults_kink = interp_kink._compute_knots(interp_kink.m_ncp, params)
        
        # The number of knots and mults might increase due to kink insertion
        # The sum of multiplicities should still be ncp + degree + 1, but the structure changes
        # The new _insert_knot_with_multiplicity ensures multiplicity is capped at degree
        assert sum(mults_kink) >= interp_kink.m_ncp + interp_kink.m_degree + 1
        assert mults_kink[0] == interp_kink.m_degree + 1
        assert mults_kink[-1] == interp_kink.m_degree + 1
        # Check if the kink parameter is present with increased multiplicity
        kink_param = params[1]
        found_kink_knot = False
        for i, k in enumerate(knots_kink):
            if math.isclose(k, kink_param, rel_tol=1e-4):
                assert mults_kink[i] >= interp_kink.m_degree # Multiplicity should be at least degree
                found_kink_knot = True
                break
        assert found_kink_knot

    def test_fit_curve_simple(self, linear_points_array):
        interp = BSplineApproxInterp(linear_points_array, n_control_points=3, degree=1)
        result = interp.fit_curve()
        assert result.curve is not None
        assert isinstance(result.curve, Geom_BSplineCurve)
        assert result.error >= 0.0
        # For linear points and degree 1, error should be very small
        assert result.error < 1e-6

        # Check if the curve interpolates the points
        for i in range(linear_points_array.Length()):
            p_expected = linear_points_array(i + 1)
            param = interp._compute_parameters(0.5)[i]
            p_actual = result.curve.Value(param)
            assert p_expected.IsEqual(p_actual, 1e-6)

    def test_fit_curve_optimal_simple(self, linear_points_array):
        interp = BSplineApproxInterp(linear_points_array, n_control_points=3, degree=1)
        result = interp.fit_curve_optimal()
        assert result.curve is not None
        assert isinstance(result.curve, Geom_BSplineCurve)
        assert result.error >= 0.0
        assert result.error < 1e-6

        # Check if the curve interpolates the points
        for i in range(linear_points_array.Length()):
            p_expected = linear_points_array(i + 1)
            # Optimal parameters might be slightly different, but should still interpolate
            # For linear, they should be the same as computed
            param = interp._compute_parameters(0.5)[i] # Re-compute for comparison
            p_actual = result.curve.Value(param)
            assert p_expected.IsEqual(p_actual, 1e-6)

    def test_fit_curve_with_interpolation(self, simple_points_array):
        interp = BSplineApproxInterp(simple_points_array, n_control_points=4, degree=3)
        interp.interpolate_point(0) # Interpolate first point
        interp.interpolate_point(3) # Interpolate last point
        
        result = interp.fit_curve()
        assert result.curve is not None
        assert result.error >= 0.0

        # Check if interpolated points are exact
        params = interp._compute_parameters(0.5)
        p0_expected = simple_points_array(1)
        p0_actual = result.curve.Value(params[0])
        assert p0_expected.IsEqual(p0_actual, 1e-6)

        p3_expected = simple_points_array(4)
        p3_actual = result.curve.Value(params[3])
        assert p3_expected.IsEqual(p3_actual, 1e-6)

    def test_fit_curve_with_kinks(self, simple_points_array):
        interp = BSplineApproxInterp(simple_points_array, n_control_points=5, degree=2)
        interp.interpolate_point(1, with_kink=True) # Kink at index 1
        
        result = interp.fit_curve()
        assert result.curve is not None
        assert result.error >= 0.0

        # Check if the kink parameter has high multiplicity in the resulting curve
        params = interp._compute_parameters(0.5)
        kink_param = params[1]
        
        found_kink_mult = False
        for i in range(1, result.curve.NbKnots() + 1):
            if math.isclose(result.curve.Knot(i), kink_param, rel_tol=1e-4):
                assert result.curve.Multiplicity(i) >= interp.m_degree
                found_kink_mult = True
                break
        assert found_kink_mult

    def test_project_on_curve(self, linear_points_array):
        # Create a simple linear B-spline curve
        poles = TColgp_Array1OfPnt(1, 2)
        poles.SetValue(1, gp_Pnt(0, 0, 0))
        poles.SetValue(2, gp_Pnt(2, 0, 0))
        knots = BSplineAlgorithms.to_array([0.0, 1.0])
        mults = BSplineAlgorithms.to_array_int([2, 2]) # Degree 1, clamped
        curve = Geom_BSplineCurve(poles, knots.Array1(), mults.Array1(), 1)

        # Test point on the curve
        pnt_on_curve = gp_Pnt(1, 0, 0)
        initial_param_on = 0.5
        interp = BSplineApproxInterp(linear_points_array, n_control_points=3, degree=1)
        res_on = interp._project_on_curve(pnt_on_curve, curve, initial_param_on)
        assert math.isclose(res_on.parameter, 0.5, rel_tol=1e-6)
        assert res_on.error < 1e-6

        # Test point off the curve
        pnt_off_curve = gp_Pnt(1, 1, 0)
        initial_param_off = 0.5
        res_off = interp._project_on_curve(pnt_off_curve, curve, initial_param_off)
        assert math.isclose(res_off.parameter, 0.5, rel_tol=1e-6) # Closest point is at 0.5
        assert math.isclose(res_off.error, 1.0, rel_tol=1e-6) # Distance to (1,0,0) is 1.0

if __name__ == "__main__":
    # pytest.main([f'{__file__}::TestBSplineApproxInterp::test_fit_curve_with_kinks', "-v"])
    pytest.main([f'{__file__}', "-v"])
